"""Intra-ecosystem Compatibility."""

__all__: list[str] = []


from typing import Any, Final, TypeAlias

from jaxtyping import Shaped
from plum import convert

import unxt as u
from dataclassish import replace

from .base import AbstractOperator
from .pipe import Pipe
from coordinax._src.custom_types import TimeBatchOrScalar
from coordinax._src.vectors.collection import KinematicSpace
from coordinax._src.vectors.d1 import CartesianPos1D
from coordinax._src.vectors.d2 import CartesianPos2D
from coordinax._src.vectors.d3 import CartesianPos3D
from coordinax._src.vectors.d4 import FourVector

# ============================================================================
# 1-Dimensional

Q1: TypeAlias = Shaped[u.AbstractQuantity, "*#batch 1"]


@AbstractOperator.__call__.dispatch
def call(self: AbstractOperator, x: Q1, /, **kwargs: Any) -> Q1:
    """Dispatch to the operator's `__call__` method.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1], "km")
    >>> q = u.Quantity([0], "km")
    >>> op(q)
    Quantity(Array([1], dtype=int32), unit='km')

    """
    # Quantity -> CartesianPos1D -> [Operator] -> Quantity
    return convert(self(CartesianPos1D.from_(x), **kwargs), u.Quantity)


@AbstractOperator.__call__.dispatch
def call(
    self: AbstractOperator, t: TimeBatchOrScalar, x: Q1, /, **kwargs: Any
) -> tuple[TimeBatchOrScalar, Q1]:
    """Dispatch to the operator's `__call__` method.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1], "km")
    >>> q = u.Quantity([0], "km")
    >>> t = u.Quantity(0, "s")
    >>> op(t, q)
    (Quantity(Array(0, dtype=int32, ...), unit='s'),
     Quantity(Array([1], dtype=int32), unit='km'))

    """
    t, vec = self(t, CartesianPos1D.from_(x), **kwargs)
    return t, convert(vec, u.Quantity)


# ============================================================================
# 2-Dimensional


Q2: TypeAlias = Shaped[u.AbstractQuantity, "*#batch 2"]


@AbstractOperator.__call__.dispatch
def call(self: AbstractOperator, x: Q2, /, **kwargs: Any) -> Q2:
    """Dispatch to the operator's `__call__` method.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> q = u.Quantity([1, 2], "m")
    >>> op = cx.ops.GalileanSpatialTranslation(u.Quantity([-1, -1], "m"))
    >>> op(q)
    Quantity(Array([0, 1], dtype=int32), unit='m')

    """
    return convert(self(CartesianPos2D.from_(x), **kwargs), u.Quantity)


@AbstractOperator.__call__.dispatch
def call(
    self: AbstractOperator, t: TimeBatchOrScalar, x: Q2, /, **kwargs: Any
) -> tuple[TimeBatchOrScalar, Q2]:
    """Dispatch to the operator's `__call__` method.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> q = u.Quantity([1, 2], "m")
    >>> t = u.Quantity(0, "s")
    >>> op = cx.ops.GalileanSpatialTranslation(u.Quantity([-1, -1], "m"))
    >>> op(t, q)
    (Quantity(Array(0, dtype=int32, ...), unit='s'),
     Quantity(Array([0, 1], dtype=int32), unit='m'))

    """
    t, vec = self(t, CartesianPos2D.from_(x), **kwargs)
    return t, convert(vec, u.Quantity)


# ============================================================================
# 3-Dimensional


Q3: TypeAlias = Shaped[u.AbstractQuantity, "*#batch 3"]


@AbstractOperator.__call__.dispatch_multi(
    (AbstractOperator, Q3),
    (Pipe, Q3),
)
def call(self: AbstractOperator, q: Q3, /, **kwargs: Any) -> Q3:
    r"""Operate on a 3D Quantity.

    `q` is the position vector. This is interpreted as a 3D CartesianVector.
    See `coordinax.CartesianPos3D` for more details.

    Returns
    -------
    x' : Quantity['length', '*#batch 3']
        The operated-upon position vector.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> shift = u.Quantity([1.0, 2.0, 3.0], "km")
    >>> op = cx.ops.GalileanSpatialTranslation(shift)

    >>> q = u.Quantity([0.0, 0, 0], "km")
    >>> op(q)
    Quantity(Array([1., 2., 3.], dtype=float32), unit='km')

    """
    cart = CartesianPos3D.from_(q)
    result = self(cart, **kwargs)
    return convert(result.vconvert(CartesianPos3D), u.Quantity)


@AbstractOperator.__call__.dispatch
def call(
    self: AbstractOperator, t: TimeBatchOrScalar, x: Q3, /, **kwargs: Any
) -> tuple[TimeBatchOrScalar, Q3]:
    """Dispatch to the operator's `__call__` method.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    We can then create a spatial translation operator:

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1, 2, 3], "km")
    >>> op
    GalileanSpatialTranslation(CartesianPos3D( ... ))

    We can then apply the operator to a position:

    >>> q = u.Quantity([1.0, 2.0, 3.0], "km")
    >>> t = u.Quantity(0.0, "Gyr")

    >>> op(t, q)
    (Quantity(Array(0., dtype=float32, ...), unit='Gyr'),
     Quantity(Array([2., 4., 6.], dtype=float32), unit='km'))

    """
    t, vec = self(t, CartesianPos3D.from_(x), **kwargs)
    return t, convert(vec, u.Quantity)


# ============================================================================
# 4-Dimensional


@AbstractOperator.__call__.dispatch
def call(self: AbstractOperator, v4: FourVector, /, **kwargs: Any) -> FourVector:
    """Dispatch to the operator's `__call__` method.

    Examples
    --------
    >>> import coordinax as cx

    We can then create a spatial translation operator:

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1, 2, 3], "km")
    >>> op
    GalileanSpatialTranslation(CartesianPos3D( ... ))

    We can then apply the operator to a position:

    >>> pos = cx.FourVector.from_([0, 1.0, 2.0, 3.0], "km")
    >>> pos
    FourVector( t=Quantity(...), q=CartesianPos3D( ... ) )

    >>> newpos = op(pos)
    >>> newpos
    FourVector( t=Quantity(...), q=CartesianPos3D( ... ) )
    >>> newpos.q.x
    Quantity(Array(2., dtype=float32), unit='km')

    Now on a VelocityBoost:

    >>> op = cx.ops.VelocityBoost.from_([1, 2, 3], "m/s")

    >>> v4 = cx.FourVector.from_([0, 0, 0, 0], "m")
    >>> newv4 = op(v4)
    >>> print(newv4)
    <FourVector: (t[m s / km], q=(x, y, z) [m])
        [0. 0. 0. 0.]>

    """
    t, q = self(v4.t, v4.q, **kwargs)
    return FourVector(t=t, q=q)


@AbstractOperator.__call__.dispatch
def call(
    self: AbstractOperator,
    x: Shaped[u.Quantity["length"], "*batch 4"],
    /,
    **kwargs: Any,
) -> Shaped[u.Quantity["length"], "*batch 4"]:
    """Dispatch to the operator's `__call__` method.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    We can then create a spatial translation operator:

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1, 2, 3], "km")
    >>> op
    GalileanSpatialTranslation(CartesianPos3D( ... ))

    We can then apply the operator to a position:

    >>> pos = u.Quantity([0, 1.0, 2.0, 3.0], "km")
    >>> pos
    Quantity(Array([0., 1., 2., 3.], dtype=float32), unit='km')

    >>> newpos = op(pos)
    >>> newpos
    Quantity(Array([0., 2., 4., 6.], dtype=float32), unit='km')

    """
    q4 = FourVector.from_(x)
    return convert(self(q4, **kwargs), u.Quantity)


# ============================================================================
# Space

# TODO: enable arbitrary keys
known_keys: Final = ("length", "speed", "acceleration")


@AbstractOperator.__call__.dispatch  # TODO: KinematicSpace[PosT] -- plum#212
def call(op: AbstractOperator, space: KinematicSpace, /, **__: Any) -> KinematicSpace:
    r"""Apply the boost to a Space.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    On positions:

    >>> op = cx.ops.GalileanRotation.from_([[0., -1, 0], [1, 0, 0], [0, 0, 1]])

    >>> x = cx.CartesianPos3D.from_([1., 2, 3], "m")
    >>> space = cx.KinematicSpace(length=x)
    >>> new_space = op(space)  # no effect
    >>> print(new_space["length"])
    <CartesianPos3D: (x, y, z) [m]
        [-2.  1.  3.]>

    On positions and velocities:

    >>> op = cx.ops.VelocityBoost.from_([1, 2, 3], "m/s")

    >>> v = cx.CartesianVel3D.from_([4., 5, 6], "m/s")
    >>> space = cx.KinematicSpace(length=x, speed=v)
    >>> space
    KinematicSpace({ 'length': CartesianPos3D( ... ), 'speed': CartesianVel3D( ... ) })

    >>> new_space = op(space)
    >>> new_space.keys()
    dict_keys(['length', 'speed'])
    >>> print(new_space["speed"])
    <CartesianVel3D: (x, y, z) [m / s]
        [5. 7. 9.]>

    """
    ks = [k for k in known_keys if k in space]
    vecs = op(*(space[k] for k in ks))
    vecs = (vecs,) if len(ks) == 1 else vecs
    return replace(space, **dict(zip(ks, vecs, strict=True)))


@AbstractOperator.__call__.dispatch  # TODO: KinematicSpace[PosT] -- plum#212
def call(
    op: AbstractOperator, t: u.Quantity, space: KinematicSpace, /, **__: Any
) -> tuple[u.Quantity, KinematicSpace]:
    """Apply the operator to a Space with a time argument.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> space = cx.KinematicSpace(length=cx.CartesianPos3D.from_([1., 2, 3], "m"))

    >>> op = cx.ops.GalileanRotation.from_([[0., -1, 0], [1, 0, 0], [0, 0, 1]])
    >>> t = u.Quantity(0.0, "s")
    >>> _, new_space = op(t, space)
    >>> print(new_space["length"])
    <CartesianPos3D: (x, y, z) [m]
        [-2.  1.  3.]>

    """
    ks = [k for k in known_keys if k in space]
    t, vecs = op(t, *(space[k] for k in ks))
    vecs = (vecs,) if len(ks) == 1 else vecs
    out = replace(space, **dict(zip(ks, vecs, strict=True)))
    return t, out



"""Built-in vector classes."""

__all__ = (
    "CartesianAcc2D",
    "CartesianPos2D",
    "CartesianVel2D",
)

import functools as ft
from typing import final
from typing_extensions import override

import equinox as eqx

import quaxed.numpy as jnp
import unxt as u

import coordinax._src.custom_types as ct
from .base import AbstractAcc2D, AbstractPos2D, AbstractVel2D
from coordinax._src.distances import BBtLength
from coordinax._src.vectors.base.cartesian import AbstractCartesian


@final
class CartesianPos2D(AbstractCartesian, AbstractPos2D):
    """Cartesian 2D Position.

    Examples
    --------
    >>> import coordinax as cx

    >>> vec = cx.vecs.CartesianPos2D.from_([1, 2], "m")
    >>> print(vec)
    <CartesianPos2D: (x, y) [m]
        [1 2]>

    """

    x: BBtLength = eqx.field(converter=u.Quantity["length"].from_)
    r"""X coordinate :math:`x \in (-\infty,+\infty)`."""

    y: BBtLength = eqx.field(converter=u.Quantity["length"].from_)
    r"""Y coordinate :math:`y \in (-\infty,+\infty)`."""


@final
class CartesianVel2D(AbstractCartesian, AbstractVel2D):
    """Cartesian 2D Velocity.

    Examples
    --------
    >>> import coordinax as cx

    >>> vec = cx.vecs.CartesianVel2D.from_([1, 2], "m/s")
    >>> print(vec)
    <CartesianVel2D: (x, y) [m / s]
        [1 2]>

    """

    x: ct.BBtSpeed = eqx.field(converter=u.Quantity["speed"].from_)
    r"""X coordinate differential :math:`\dot{x} \in (-\infty,+\infty)`."""

    y: ct.BBtSpeed = eqx.field(converter=u.Quantity["speed"].from_)
    r"""Y coordinate differential :math:`\dot{y} \in (-\infty,+\infty)`."""

    @override
    def norm(self, _: AbstractPos2D | None = None, /) -> ct.BBtSpeed:
        """Return the norm of the vector.

        Examples
        --------
        >>> import coordinax as cx
        >>> v = cx.vecs.CartesianVel2D.from_([3, 4], "km/s")
        >>> v.norm()
        Quantity(Array(5., dtype=float32), unit='km / s')

        """
        return jnp.sqrt(self.x**2 + self.y**2)


@final
class CartesianAcc2D(AbstractCartesian, AbstractAcc2D):
    """Cartesian Acceleration 3D.

    Examples
    --------
    >>> import coordinax as cx

    >>> vec = cx.vecs.CartesianAcc2D.from_([1, 2], "m/s2")
    >>> print(vec)
    <CartesianAcc2D: (x, y) [m / s2]
        [1 2]>

    """

    x: ct.BBtAcc = eqx.field(converter=u.Quantity["acceleration"].from_)
    r"""X coordinate acceleration :math:`\frac{d^2 x}{dt^2} \in (-\infty,+\infty)`."""

    y: ct.BBtAcc = eqx.field(converter=u.Quantity["acceleration"].from_)
    r"""Y coordinate acceleration :math:`\frac{d^2 y}{dt^2} \in (-\infty,+\infty)`."""

    @override
    @ft.partial(eqx.filter_jit, inline=True)
    def norm(self, _: AbstractVel2D | None = None, /) -> ct.BBtAcc:  # type: ignore[misc, override]
        """Return the norm of the vector.

        Examples
        --------
        >>> import coordinax as cx
        >>> v = cx.vecs.CartesianAcc2D.from_([3, 4], "km/s2")
        >>> v.norm()
        Quantity(Array(5., dtype=float32), unit='km / s2')

        """
        return jnp.sqrt(self.x**2 + self.y**2)



"""4-dimensional."""

__all__ = (
    # Base
    "AbstractPos4D",
    # Spacetime
    "FourVector",
)

from .base import AbstractPos4D
from .spacetime import FourVector

# Register by importing
# isort: split
from . import (
    register_convert,  # noqa: F401
    register_primitives,  # noqa: F401
    register_vectorapi,  # noqa: F401
)



"""Doctest configuration."""

from collections.abc import Callable, Iterable, Sequence
from doctest import ELLIPSIS, NORMALIZE_WHITESPACE

from sybil import Document, Region, Sybil
from sybil.parsers import myst, rest

optionflags = ELLIPSIS | NORMALIZE_WHITESPACE

parsers: Sequence[Callable[[Document], Iterable[Region]]] = [
    myst.DocTestDirectiveParser(optionflags=optionflags),
    myst.PythonCodeBlockParser(doctest_optionflags=optionflags),
    myst.SkipParser(),
]

docs = Sybil(parsers=parsers, patterns=["*.md"])
python = Sybil(  # TODO: use myst for doctests
    parsers=[*parsers, rest.DocTestParser(optionflags=optionflags), rest.SkipParser()],
    patterns=["*.py"],
)


pytest_collect_file = (docs + python).pytest()



"""Representation of coordinates in different systems."""

__all__ = ("AbstractVectorLike", "is_vectorlike")

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, NoReturn, TypeGuard, TypeVar, cast

import jax
import quax_blocks
import wadler_lindig as wl
from jaxtyping import DTypeLike
from plum import dispatch
from quax import ArrayValue

import quaxed.numpy as jnp
import unxt as u
from dataclassish import field_items, field_values, replace

from .flags import AttrFilter
from coordinax._src.vectors.api import vconvert
from coordinax._src.vectors.utils import full_shaped

if TYPE_CHECKING:
    from typing import Self

VT = TypeVar("VT", bound="AbstractVectorLike")


class AbstractVectorLike(
    ArrayValue,
    quax_blocks.LaxBinaryOpsMixin[Any, Any],  # TODO: type annotation
    quax_blocks.LaxRoundMixin["AbstractVectorLike"],
    quax_blocks.LaxUnaryMixin[Any],
):
    """Base class for all vector-like types.

    A vector is a collection of components that can be represented in different
    coordinate systems. This class provides a common interface for all
    vector-like types, which includes vectors but also other types like
    collections of vectors that share some properties and methods.

    Methods
    -------
    vconvert
        Convert the vector(s) to another type.
        For example, a Cartesian position vector can be converted to a
        spherical position vector.
    uconvert
        Convert the vector(s) to a different unit system.
        For example, a Cartesian position vector in meters can be converted to
        kilometers.

    astype
        Cast the fields of the vector to a new dtype.
    copy
        Return a copy of the vector.
    flatten
        Flatten the fields of the vector.
    ravel
        Ravel the fields of the vector.
    reshape
        Reshape the fields of the vector.
    round
        Round the fields of the vector to a given number of decimals.
    to_device
        Move the fields of the vector to a new device.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    Scalar vectors have length 0:

    >>> vec = cx.vecs.CartesianPos1D.from_([1], "m")
    >>> len(vec)
    0

    Vectors with certain lengths:

    >>> vec = cx.vecs.CartesianPos1D(u.Quantity([1], "m"))
    >>> len(vec)
    1

    >>> vec = cx.vecs.CartesianPos1D(u.Quantity([1, 2], "m"))
    >>> len(vec)
    2

    """

    # ---------------------------------
    # Constructors

    @classmethod
    @dispatch.abstract
    def from_(
        cls: "type[AbstractVectorLike]", *args: Any, **kwargs: Any
    ) -> "AbstractVectorLike":
        """Create a vector-like object from arguments.

        Examples
        --------
        >>> import coordinax.vecs as cxv

        >>> q = cxv.CartesianPos3D.from_([1, 2, 3], "m")
        >>> print(q)
        <CartesianPos3D: (x, y, z) [m]
            [1 2 3]>

        >>> v = cxv.CartesianVel3D.from_([1, 2, 3], "m/s")
        >>> print(v)
        <CartesianVel3D: (x, y, z) [m / s]
            [1 2 3]>

        >>> space = cxv.KinematicSpace.from_(q)
        >>> print(space)
        KinematicSpace({
         'length': <CartesianPos3D: (x, y, z) [m]
                       [1 2 3]>
        })

        """
        raise NotImplementedError  # pragma: no cover

    # ===============================================================
    # Vector API

    @dispatch
    def vconvert(
        self: "AbstractVectorLike", target: type, *args: Any, **kwargs: Any
    ) -> "AbstractVectorLike":
        """Represent the vector as another type.

        This just forwards to `coordinax.vconvert`.

        Parameters
        ----------
        target : type[`coordinax.AbstractVectorLike`]
            The type to represent the vector as.
        *args, **kwargs
            Extra arguments. These are passed to `coordinax.vconvert` and
            might be used, depending on the dispatched method. E.g. for
            transforming an acceleration, generally the first argument is the
            velocity (`coordinax.AbstractVel`) followed by the position
            (`coordinax.AbstractPos`) at which the acceleration is defined. In
            general this is a required argument, though it is not for
            Cartesian-to-Cartesian transforms -- see
            https://en.wikipedia.org/wiki/Tensors_in_curvilinear_coordinates for
            more information.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax.vecs as cxv

        Transforming a Position:

        >>> q_cart = cxv.CartesianPos3D.from_([1, 2, 3], "m")

        >>> q_sph = q_cart.vconvert(cxv.SphericalPos)
        >>> print(q_sph)
        <SphericalPos: (r[m], theta[rad], phi[rad])
            [3.742 0.641 1.107]>

        >>> q_ps = q_cart.vconvert(cxv.ProlateSpheroidalPos, Delta=u.Quantity(1.5, "m"))
        >>> print(q_ps)
        <ProlateSpheroidalPos: (mu[m2], nu[m2], phi[rad])
         Delta=Quantity(1.5, unit='m')
            [14.89   1.36   1.107]>

        >>> print((q_ps.vconvert(cxv.CartesianPos3D) - q_cart).round(3))
        <CartesianPos3D: (x, y, z) [m]
            [-0.  0.  0.]>

        Transforming a Velocity:

        >>> v_cart = cxv.CartesianVel3D.from_([1, 2, 3], "m/s")
        >>> v_sph = v_cart.vconvert(cxv.SphericalVel, q_cart)
        >>> print(v_sph)
        <SphericalVel: (r[m / s], theta[rad / s], phi[rad / s])
            [ 3.742e+00 -8.941e-08  0.000e+00]>

        Transforming an Acceleration:

        >>> a_cart = cxv.CartesianAcc3D.from_([7, 8, 9], "m/s2")
        >>> a_sph = a_cart.vconvert(cxv.SphericalAcc, v_cart, q_cart)
        >>> print(a_sph)
        <SphericalAcc: (r[m / s2], theta[rad / s2], phi[rad / s2])
            [13.363  0.767 -1.2  ]>

        """
        return vconvert(target, self, *args, **kwargs)

    # ===============================================================
    # Quantity API

    @dispatch(precedence=-1)
    def uconvert(self, *args: Any, **kwargs: Any) -> "AbstractVectorLike":
        """Convert the vector to the given units.

        This just forwards to `unxt.uconvert`, reversing the order of the
        arguments to match the `unxt` API.

        Examples
        --------
        >>> import coordinax as cx

        >>> vec = cx.vecs.CartesianPos3D.from_([1, 2, 3], "km")
        >>> vec.uconvert({"length": "km"})
        CartesianPos3D(
            x=Quantity(1, unit='km'), y=Quantity(2, unit='km'), z=Quantity(3, unit='km')
        )

        """
        return u.uconvert(*args, self, **kwargs)

    @dispatch
    def uconvert(self, usys: Any, /) -> "AbstractVectorLike":
        """Convert the vector to the given units.

        Parameters
        ----------
        usys
            The units to convert to according to the physical type of the
            components. This is passed to [`unxt.unitsystem`][].

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> usys = u.unitsystem("m", "s", "kg", "rad")

        >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "km")

        >>> print(vec.uconvert(usys))
        <CartesianPos3D: (x, y, z) [m]
            [1000. 2000. 3000.]>

        >>> print(vec.uconvert("galactic"))
        <CartesianPos3D: (x, y, z) [kpc]
            [3.241e-17 6.482e-17 9.722e-17]>

        """
        return u.uconvert(usys, self)

    # ===============================================================
    # Quax API

    def materialise(self) -> NoReturn:
        """Materialise the vector for `quax`.

        Examples
        --------
        >>> import coordinax as cx
        >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "m")

        >>> try: vec.materialise()
        ... except RuntimeError as e: print(e)
        Refusing to materialise `CartesianPos3D`.

        """
        msg = f"Refusing to materialise `{type(self).__name__}`."
        raise RuntimeError(msg)

    @property
    def shape(self) -> tuple[int, ...]:
        """Return the shape of the vector.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        We can get the shape of a vector:

        >>> vec = cx.vecs.CartesianPos1D(x=u.Quantity([1, 2], "m"))
        >>> vec.shape
        (2,)

        >>> vec = cx.vecs.CartesianPos1D(x=u.Quantity([[1, 2], [3, 4]], "m"))
        >>> vec.shape
        (2, 2)

        ``shape`` is calculated from the broadcasted shape. We can
        see this by creating a 2D vector in which the components have
        different shapes:

        >>> vec = cx.vecs.CartesianPos2D(x=u.Quantity([[1, 2], [3, 4]], "m"),
        ...                              y=u.Quantity(0, "m"))
        >>> vec.shape
        (2, 2)

        >>> space = cx.KinematicSpace(length=vec)
        >>> space.shape
        (2, 2)

        """
        shapes = [v.shape for v in field_values(AttrFilter, self)]
        return jnp.broadcast_shapes(*shapes)

    # ===============================================================
    # Array API

    def __array_namespace__(self) -> Any:
        """Return the array API namespace.

        Here we return the `quaxed.numpy` module, which is a drop-in replacement
        for `jax.numpy`, but allows for array-ish objects to be used in place of
        `jax` arrays. See `quax` for more information.

        Examples
        --------
        >>> import coordinax as cx
        >>> vec = cx.vecs.CartesianPos2D.from_([3, 4], "m")
        >>> ns = vec.__array_namespace__()
        >>> ns
        <module 'quaxed.numpy' from ...>

        >>> ns.multiply(vec, 2)
        CartesianPos2D(x=Quantity(6, unit='m'), y=Quantity(8, unit='m'))

        """
        return jnp

    # ---------------------------------
    # comparison operators

    # TODO: use quax_blocks.LaxEqMixin
    def __eq__(self: "AbstractVectorLike", other: object) -> Any:
        """Check if the vector is equal to another object.

        Examples
        --------
        >>> import quaxed.numpy as jnp
        >>> import unxt as u
        >>> import coordinax as cx

        We can compare non-vector objects:

        >>> vec = cx.vecs.CartesianPos1D(u.Quantity([1, 2], "m"))
        >>> vec == 2
        False

        And positions.

        >>> q = cx.vecs.CylindricalPos(rho=u.Quantity([1.0, 2.0], "kpc"),
        ...                            phi=u.Quantity([0.0, 0.2], "rad"),
        ...                            z=u.Quantity(0.0, "kpc"))
        >>> q == q
        Array([ True,  True], dtype=bool)

        Showing the broadcasting, then element-wise comparison of two vectors:

        >>> vec1 = cx.CartesianPos3D.from_([[1, 2, 3], [1, 2, 4]], "m")
        >>> vec2 = cx.CartesianPos3D.from_([1, 2, 3], "m")
        >>> jnp.equal(vec1, vec2)
        Array([ True, False], dtype=bool)

        Showing the change of representation:

        >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "m")
        >>> vec1 = vec.vconvert(cx.SphericalPos)
        >>> vec2 = vec.vconvert(cx.vecs.MathSphericalPos)
        >>> jnp.equal(vec1, vec2)
        Array(True, dtype=bool)

        Quick run-through of each dimensionality:

        >>> vec1 = cx.vecs.CartesianPos1D.from_([1], "m")
        >>> vec2 = cx.vecs.RadialPos.from_([1], "m")
        >>> jnp.equal(vec1, vec2)
        Array(True, dtype=bool)

        >>> vec1 = cx.vecs.CartesianPos2D.from_([2, 0], "m")
        >>> vec2 = cx.vecs.PolarPos(r=u.Quantity(2, "m"), phi=u.Quantity(0, "rad"))
        >>> jnp.equal(vec1, vec2)
        Array(True, dtype=bool)

        Now we show velocities and accelerations:

        >>> vel1 = cx.vecs.CartesianVel1D(u.Quantity([1, 2, 3], "km/s"))
        >>> vel2 = cx.vecs.CartesianVel1D(u.Quantity([1, 0, 3], "km/s"))
        >>> jnp.equal(vel1, vel2)
        Array([ True,  False,  True], dtype=bool)
        >>> vel1 == vel2
        Array([ True, False,  True], dtype=bool)

        >>> acc1 = cx.vecs.CartesianAcc1D(u.Quantity([1, 2, 3], "km/s2"))
        >>> acc2 = cx.vecs.CartesianAcc1D(u.Quantity([1, 0, 3], "km/s2"))
        >>> jnp.equal(acc1, acc2)
        Array([ True,  False,  True], dtype=bool)
        >>> acc1 == acc2
        Array([ True, False,  True], dtype=bool)

        >>> vel1 = cx.vecs.RadialVel(u.Quantity([1, 2, 3], "km/s"))
        >>> vel2 = cx.vecs.RadialVel(u.Quantity([1, 0, 3], "km/s"))
        >>> jnp.equal(vel1, vel2)
        Array([ True,  False,  True], dtype=bool)
        >>> vel1 == vel2
        Array([ True, False,  True], dtype=bool)

        >>> acc1 = cx.vecs.RadialAcc(u.Quantity([1, 2, 3], "km/s2"))
        >>> acc2 = cx.vecs.RadialAcc(u.Quantity([1, 0, 3], "km/s2"))
        >>> jnp.equal(acc1, acc2)
        Array([ True,  False,  True], dtype=bool)
        >>> acc1 == acc2
        Array([ True, False,  True], dtype=bool)

        >>> vel1 = cx.vecs.CartesianVel2D.from_([[1, 3], [2, 4]], "km/s")
        >>> vel2 = cx.vecs.CartesianVel2D.from_([[1, 3], [0, 4]], "km/s")
        >>> vel1.x
        Quantity(Array([1, 2], dtype=int32), unit='km / s')
        >>> jnp.equal(vel1, vel2)
        Array([ True, False], dtype=bool)
        >>> vel1 == vel2
        Array([ True, False], dtype=bool)

        >>> acc1 = cx.vecs.CartesianAcc2D.from_([[1, 3], [2, 4]], "km/s2")
        >>> acc2 = cx.vecs.CartesianAcc2D.from_([[1, 3], [0, 4]], "km/s2")
        >>> acc1.x
        Quantity(Array([1, 2], dtype=int32), unit='km / s2')
        >>> jnp.equal(acc1, acc2)
        Array([ True, False], dtype=bool)
        >>> acc1 == acc2
        Array([ True, False], dtype=bool)

        >>> vel1 = cx.CartesianVel3D.from_([[1, 2, 3], [4, 5, 6]], "km/s")
        >>> vel2 = cx.CartesianVel3D.from_([[1, 2, 3], [4, 5, 0]], "km/s")
        >>> vel1.x
        Quantity(Array([1, 4], dtype=int32), unit='km / s')
        >>> jnp.equal(vel1, vel2)
        Array([ True, False], dtype=bool)
        >>> vel1 == vel2
        Array([ True, False], dtype=bool)

        """
        if type(other) is not type(self):
            return NotImplemented

        return jnp.equal(self, cast("AbstractVectorLike", other))

    # ---------------------------------
    # methods

    def __complex__(self) -> NoReturn:
        """Convert a zero-dimensional object to a Python complex object."""
        raise NotImplementedError  # pragma: no cover

    # TODO: .__dlpack__, __dlpack_device__

    def __float__(self) -> NoReturn:
        """Convert a zero-dimensional object to a Python float object."""
        raise NotImplementedError  # pragma: no cover

    def __getitem__(self, index: Any) -> "Self":
        """Return a new object with the given slice applied.

        Parameters
        ----------
        index
            The slice to apply.

        Examples
        --------
        We assume the following imports:

        >>> import unxt as u
        >>> import coordinax as cx

        We can slice a vector:

        >>> vec = cx.vecs.CartesianPos2D(x=u.Quantity([[1, 2], [3, 4]], "m"),
        ...                              y=u.Quantity(0, "m"))
        >>> vec[0].x
        Quantity(Array([1, 2], dtype=int32), unit='m')

        """
        full = full_shaped(self)  # TODO: detect if need to make a full-shaped copy
        return replace(full, **{k: v[index] for k, v in field_items(AttrFilter, full)})

    def __index__(self) -> NoReturn:
        """Convert the vector to an integer index."""
        raise NotImplementedError  # pragma: no cover

    def __int__(self) -> NoReturn:
        """Convert the vector to an integer."""
        raise NotImplementedError  # pragma: no cover

    def __setitem__(self, k: Any, v: Any) -> NoReturn:
        """Fail to set an item in the vector.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        We can't set an item in a vector:

        >>> vec = cx.vecs.CartesianPos2D.from_([[1, 2], [3, 4]], "m")
        >>> try: vec[0] = u.Quantity(1, "m")
        ... except TypeError as e: print(e)
        CartesianPos2D is immutable.

        """
        msg = f"{type(self).__name__} is immutable."
        raise TypeError(msg)

    # ===============================================================
    # JAX API

    # TODO: repeat(), round(), sort(), squeeze(), swapaxes(), transpose(),
    # view() addressable_shards, at, committed, globarl_shards,
    # is_fully_addressable, is_fully_replcated, nbytes, sharding

    @dispatch
    def astype(
        self: "AbstractVectorLike", dtype: Any, /, **kwargs: Any
    ) -> "AbstractVectorLike":
        """Cast the vector to a new dtype.

        Examples
        --------
        >>> import quaxed.numpy as jnp
        >>> import unxt as u
        >>> import coordinax as cx

        We can cast a vector to a new dtype:

        >>> vec = cx.vecs.CartesianPos1D(u.Quantity([1, 2], "m"))
        >>> vec.astype(jnp.float32)
        CartesianPos1D(x=Quantity([1., 2.], unit='m'))

        >>> jnp.astype(vec, jnp.float32)
        CartesianPos1D(x=Quantity([1., 2.], unit='m'))

        """
        return replace(
            self, **{k: v.astype(dtype, **kwargs) for k, v in field_items(self)}
        )

    @dispatch
    def astype(
        self: "AbstractVectorLike", dtypes: Mapping[str, DTypeLike], /, **kwargs: Any
    ) -> "AbstractVectorLike":
        """Cast the vector to a new dtype.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        We can cast a vector to a new dtype:

        >>> vec = cx.vecs.CartesianPos1D(u.Quantity([1, 2], "m"))
        >>> vec
        CartesianPos1D(x=Quantity([1, 2], unit='m'))

        >>> vec.astype({"x": jnp.float32})
        CartesianPos1D(x=Quantity([1., 2.], unit='m'))

        """
        return replace(
            self,
            **{
                k: (v.astype(dtypes[k], **kwargs) if k in dtypes else v)
                for k, v in field_items(self)
            },
        )

    # -------------------------------

    def copy(self) -> "Self":
        """Return a copy of the vector.

        Examples
        --------
        >>> import coordinax as cx

        >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "m")
        >>> print(vec.copy())
        <CartesianPos3D: (x, y, z) [m]
            [1 2 3]>

        """
        return replace(self)  # TODO: should .copy be called on the components?

    def flatten(self) -> "Self":
        """Flatten the vector.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> vec = cx.vecs.CartesianPos2D(x=u.Quantity([[1, 2], [3, 4]], "m"),
        ...                              y=u.Quantity(0, "m"))
        >>> vec.shape
        (2, 2)

        >>> vec.flatten().shape
        (4,)

        """
        return replace(
            self, **{k: v.flatten() for k, v in field_items(AttrFilter, self)}
        )

    def ravel(self) -> "Self":
        """Ravel the vector.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> vec = cx.vecs.CartesianPos2D(x=u.Quantity([[1, 2], [3, 4]], "m"),
        ...                              y=u.Quantity(0, "m"))
        >>> vec.shape
        (2, 2)

        >>> vec.ravel().shape
        (4,)

        """
        return replace(self, **{k: v.ravel() for k, v in field_items(AttrFilter, self)})

    def reshape(self, *shape: Any, order: str = "C") -> "Self":
        """Reshape the components of the vector.

        Parameters
        ----------
        *shape
            The new shape.
        order
            The order to use for the reshape.

        Examples
        --------
        We assume the following imports:

        >>> import unxt as u
        >>> import coordinax as cx

        We can reshape a vector:

        >>> vec = cx.vecs.CartesianPos2D(x=u.Quantity([[1, 2], [3, 4]], "m"),
        ...                              y=u.Quantity(0, "m"))

        >>> vec.reshape(4)
        CartesianPos2D(x=Quantity([1, 2, 3, 4], unit='m'),
                       y=Quantity([0, 0, 0, 0], unit='m'))

        """
        # TODO: enable not needing to make a full-shaped copy
        full = full_shaped(self)
        changes = {
            k: v.reshape(*shape, order=order) for k, v in field_items(AttrFilter, full)
        }
        return replace(self, **changes)

    def round(self, decimals: int = 0) -> "Self":
        """Round the components of the vector.

        Parameters
        ----------
        decimals
            The number of decimals to round to.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        We can round a vector:

        >>> vec = cx.vecs.CartesianPos2D.from_([[1.1, 2.2], [3.3, 4.4]], "m")
        >>> vec.round(0)
        CartesianPos2D(x=Quantity([1., 3.], unit='m'), y=Quantity([2., 4.], unit='m'))

        """
        changes = {k: v.round(decimals) for k, v in field_items(AttrFilter, self)}
        return replace(self, **changes)

    def to_device(self, device: None | jax.Device = None) -> "Self":
        """Move the vector to a new device.

        Examples
        --------
        >>> from jax import devices
        >>> import unxt as u
        >>> import coordinax as cx

        We can move a vector to a new device:

        >>> vec = cx.vecs.CartesianPos1D(u.Quantity([1, 2], "m"))
        >>> vec.to_device(devices()[0])
        CartesianPos1D(x=Quantity([1, 2], unit='m'))

        """
        changes = {
            k: v.to_device(device)
            for k, v in field_items(self)
            if hasattr(v, "to_device")
        }
        return replace(self, **changes)

    # ===============================================================
    # Python API

    def __hash__(self) -> int:
        """Return the hash of the vector.

        This is the hash of the fields, however since jax arrays are
         not hashable this will generally raise an exception.
        Defining the `__hash__` method is required for the vector to
        be considered immutable, e.g. by `dataclasses.dataclass`.

        Examples
        --------
        >>> import coordinax as cx
        >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "m")
        >>> try:
        ...     hash(vec)
        ... except TypeError as e:
        ...     print(e)
        unhashable type: 'jaxlib...ArrayImpl'

        """
        return hash(tuple(field_items(self)))

    def __repr__(self) -> str:
        """Return a string representation of the vector.

        This uses the `eqxuinox.tree_pformat` function to format the vector,
        which internally uses the `wadler_lindig` algorithm to format the string
        representation of the vector.

        """
        return wl.pformat(self, vector_form=False, short_arrays="compact")

    def __str__(self) -> str:
        """Return a string representation of the vector-like object."""
        return wl.pformat(self, vector_form=True, precision=3)


def is_vectorlike(obj: Any, /) -> TypeGuard[AbstractVectorLike]:
    """Check if the object is a `AbstractVectorLike` object.

    Examples
    --------
    >>> import coordinax as cx

    >>> vec = cx.vecs.CartesianPos3D.from_([1, 2, 3], "m")
    >>> cx.vecs.is_vectorlike(vec)
    True

    >>> space = cx.vecs.KinematicSpace.from_(vec)
    >>> cx.vecs.is_vectorlike(space)
    True

    >>> coord = cx.Coordinate.from_(vec, cx.frames.ICRS())
    >>> cx.vecs.is_vectorlike(coord)
    True

    >>> cx.vecs.is_vectorlike(42)
    False

    """
    return isinstance(obj, AbstractVectorLike)



"""Astronomy reference frames."""

__all__ = ("AbstractSpaceFrame",)


import coordinax.frames as cxf


class AbstractSpaceFrame(cxf.AbstractReferenceFrame):
    """ABC for space-related reference frames."""



"""Built-in vector classes."""

__all__ = ("PolarAcc", "PolarPos", "PolarVel")

from typing import final
from typing_extensions import override

import equinox as eqx

import unxt as u
from dataclassish.converters import Unless

import coordinax._src.custom_types as ct
from .base import AbstractAcc2D, AbstractPos2D, AbstractVel2D
from coordinax._src.angles import BatchableAngle
from coordinax._src.distances import AbstractDistance, BatchableDistance, Distance
from coordinax._src.vectors.converters import converter_azimuth_to_range


@final
class PolarPos(AbstractPos2D):
    r"""Polar vector representation."""

    r: BatchableDistance = eqx.field(converter=Unless(AbstractDistance, Distance.from_))
    r"""Radial distance :math:`r \in [0,+\infty)`."""

    phi: BatchableAngle = eqx.field(
        converter=Unless(
            u.Angle, lambda x: converter_azimuth_to_range(u.Angle.from_(x))
        )
    )
    r"""Polar angle, generally :math:`\phi \in [0,2\pi)`.

    We use the symbol `phi` to adhere to the ISO standard 31-11.
    """

    @override
    def norm(self) -> BatchableDistance:
        """Return the norm of the vector.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> vec = cx.vecs.PolarPos(r=u.Quantity(1, "m"),
        ...                        phi=u.Quantity(90, "deg"))
        >>> vec.norm()
        Distance(Array(1, dtype=int32, ...), unit='m')

        """
        return self.r


@final
class PolarVel(AbstractVel2D):
    """Polar Velocity.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vev = cx.vecs.PolarVel(r=u.Quantity(1, "m/s"), phi=u.Quantity(90, "deg/s"))
    >>> print(vev)
    <PolarVel: (r[m / s], phi[deg / s])
        [ 1 90]>

    """

    r: ct.BBtSpeed = eqx.field(converter=u.Quantity["speed"].from_)
    r"""Radial speed :math:`dr/dt \in [-\infty,+\infty]`."""

    phi: ct.BBtAngularSpeed = eqx.field(converter=u.Quantity["angular speed"].from_)
    r"""Polar angular speed :math:`d\phi/dt \in [-\infty,+\infty]`."""


#####################################################################


@final
class PolarAcc(AbstractAcc2D):
    """Polar acceleration.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> acc = cx.vecs.PolarAcc(r=u.Quantity(1, "m/s2"),
    ...                        phi=u.Quantity(3, "deg/s2"))
    >>> print(acc)
    <PolarAcc: (r[m / s2], phi[deg / s2])
        [1 3]>

    """

    r: ct.BBtAcc = eqx.field(converter=u.Quantity["acceleration"].from_)
    r"""Radial acceleration :math:`d^2r/dt^2 \in [-\infty,+\infty]`."""

    phi: ct.BBtAngularAcc = eqx.field(
        converter=u.Quantity["angular acceleration"].from_
    )
    r"""Polar angular acceleration :math:`d^2\phi/dt^2 \in [-\infty,+\infty]`."""



"""Representation of coordinates in different systems."""

__all__ = ("AbstractAcc2D", "AbstractPos2D", "AbstractVel2D")


from coordinax._src.vectors.base_acc import AbstractAcc
from coordinax._src.vectors.base_pos import AbstractPos
from coordinax._src.vectors.base_vel import AbstractVel


class AbstractPos2D(AbstractPos):
    """Abstract representation of 2D coordinates in different systems."""

    @classmethod
    def _dimensionality(cls) -> int:
        """Dimensionality of the vector.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.vecs.CartesianPos2D._dimensionality()
        2

        """
        return 2


class AbstractVel2D(AbstractVel):
    """Abstract representation of 2D vector differentials."""

    @classmethod
    def _dimensionality(cls) -> int:
        """Dimensionality of the vector.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.vecs.CartesianVel2D._dimensionality()
        2

        """
        return 2


class AbstractAcc2D(AbstractAcc):
    """Abstract representation of 2D vector accelerations."""

    @classmethod
    def _dimensionality(cls) -> int:
        """Dimensionality of the vector.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.vecs.CartesianAcc2D._dimensionality()
        2

        """
        return 2



"""Representation of coordinates in different systems."""

__all__: tuple[str, ...] = ()

from typing import NoReturn

from plum import dispatch

import quaxed.numpy as jnp
from unxt.quantity import AbstractQuantity

from .base import AbstractAccND, AbstractPosND, AbstractVelND
from .cartesian import CartesianAccND, CartesianPosND, CartesianVelND
from .poincare import PoincarePolarVector

###############################################################################


@dispatch
def vector(
    cls: type[CartesianPosND] | type[CartesianVelND] | type[CartesianAccND],
    x: AbstractQuantity,
    /,
) -> CartesianPosND | CartesianVelND | CartesianAccND:
    """Construct an N-dimensional acceleration.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    1D vector:

    >>> cx.vecs.CartesianPosND.from_(u.Quantity(1, "km"))
    CartesianPosND(q=Quantity([1], unit='km'))

    >>> cx.vecs.CartesianPosND.from_(u.Quantity([1], "km"))
    CartesianPosND(q=Quantity([1], unit='km'))

    >>> cx.vecs.CartesianVelND.from_(u.Quantity(1, "km/s"))
    CartesianVelND(q=Quantity([1], unit='km / s'))

    >>> cx.vecs.CartesianVelND.from_(u.Quantity([1], "km/s"))
    CartesianVelND(q=Quantity([1], unit='km / s'))

    >>> cx.vecs.CartesianAccND.from_(u.Quantity(1, "km/s2"))
    CartesianAccND(q=Quantity([1], unit='km / s2'))

    >>> cx.vecs.CartesianAccND.from_(u.Quantity([1], "km/s2"))
    CartesianAccND(q=Quantity([1], unit='km / s2'))

    2D vector:

    >>> cx.vecs.CartesianPosND.from_(u.Quantity([1, 2], "km"))
    CartesianPosND(q=Quantity([1, 2], unit='km'))

    >>> cx.vecs.CartesianVelND.from_(u.Quantity([1, 2], "km/s"))
    CartesianVelND(q=Quantity([1, 2], unit='km / s'))

    >>> cx.vecs.CartesianAccND.from_(u.Quantity([1, 2], "km/s2"))
    CartesianAccND(q=Quantity([1, 2], unit='km / s2'))

    3D vector:

    >>> cx.vecs.CartesianPosND.from_(u.Quantity([1, 2, 3], "km"))
    CartesianPosND(q=Quantity([1, 2, 3], unit='km'))

    >>> cx.vecs.CartesianVelND.from_(u.Quantity([1, 2, 3], "km/s"))
    CartesianVelND(q=Quantity([1, 2, 3], unit='km / s'))

    >>> cx.vecs.CartesianAccND.from_(u.Quantity([1, 2, 3], "km/s2"))
    CartesianAccND(q=Quantity([1, 2, 3], unit='km / s2'))

    4D vector:

    >>> cx.vecs.CartesianPosND.from_(u.Quantity([1, 2, 3, 4], "km"))
    CartesianPosND(q=Quantity([1, 2, 3, 4], unit='km'))

    >>> cx.vecs.CartesianVelND.from_(u.Quantity([1, 2, 3, 4], "km/s"))
    CartesianVelND(q=Quantity([1, 2, 3, 4], unit='km / s'))

    >>> cx.vecs.CartesianAccND.from_(u.Quantity([1, 2, 3, 4], "km/s2"))
    CartesianAccND(q=Quantity([1, 2, 3, 4], unit='km / s2'))

    """
    return cls(jnp.atleast_1d(x))


###############################################################################
# Corresponding Cartesian class


@dispatch
def cartesian_vector_type(
    obj: type[AbstractPosND] | AbstractPosND, /
) -> type[CartesianPosND]:
    """Return the corresponding Cartesian vector class."""
    return CartesianPosND


@dispatch
def cartesian_vector_type(
    obj: type[AbstractVelND] | AbstractVelND, /
) -> type[CartesianVelND]:
    """Return the corresponding Cartesian vector class."""
    return CartesianVelND


@dispatch
def cartesian_vector_type(
    obj: type[AbstractAccND] | AbstractAccND, /
) -> type[CartesianAccND]:
    """Return the corresponding Cartesian vector class."""
    return CartesianAccND


@dispatch
def cartesian_vector_type(
    obj: type[PoincarePolarVector] | PoincarePolarVector, /
) -> NoReturn:
    """Return the corresponding Cartesian vector class.

    Examples
    --------
    >>> import coordinax as cx
    >>> try: cx.vecs.cartesian_vector_type(cx.vecs.PoincarePolarVector)
    ... except NotImplementedError as e:
    ...     print(e)
    PoincarePolarVector does not have a corresponding Cartesian class.

    """
    msg = "PoincarePolarVector does not have a corresponding Cartesian class."
    raise NotImplementedError(msg)


###############################################################################
# Corresponding time derivative classes

# -----------------------------------------------
# Position -> Velocity


@dispatch
def time_derivative_vector_type(
    obj: type[CartesianPosND] | CartesianPosND, /
) -> type[CartesianVelND]:
    """Return the corresponding time derivative class."""
    return CartesianVelND


# -----------------------------------------------
# Velocity -> Position


@dispatch
def time_antiderivative_vector_type(
    obj: type[CartesianVelND] | CartesianVelND, /
) -> type[CartesianPosND]:
    """Return the corresponding time antiderivative class."""
    return CartesianPosND


# -----------------------------------------------
# Velocity -> Acceleration


@dispatch
def time_derivative_vector_type(
    obj: type[CartesianVelND] | CartesianVelND, /
) -> type[CartesianAccND]:
    """Return the corresponding time derivative class."""
    return CartesianAccND


# -----------------------------------------------
# Acceleration -> Velocity


@dispatch
def time_antiderivative_vector_type(
    obj: type[CartesianAccND] | CartesianAccND, /
) -> type[CartesianVelND]:
    """Return the corresponding time antiderivative class."""
    return CartesianVelND



"""Coordinax vectors.

Note that this module is private. Users should use the public API.

This module depends on the following modules:

- utils & typing
- angle
- distance


"""

__all__: tuple[str, ...] = ()


# Register vector transformations, functions, etc.
from . import funcs, register_vconvert  # noqa: F401

# Interoperability
# isort: split
from . import register_jax, register_unxt  # noqa: F401



"""Representation of coordinates in different systems."""

__all__ = ("PosT",)

from typing_extensions import TypeVar

from .core import AbstractPos

PosT = TypeVar("PosT", bound=AbstractPos, default=AbstractPos)



"""Poincare."""

__all__ = ("PoincarePolarVector",)

from typing import final

import equinox as eqx
from jaxtyping import Real

import unxt as u

import coordinax._src.custom_types as ct
from coordinax._src.distances import BBtLength
from coordinax._src.vectors import dims
from coordinax._src.vectors.base_pos import AbstractPos


@final
class PoincarePolarVector(AbstractPos):  # TODO: better name
    """Poincare vector + differential."""

    rho: BBtLength = eqx.field(converter=u.Quantity["length"].from_)
    r"""Cylindrical radial distance :math:`\rho \in [0,+\infty)`."""

    pp_phi: Real[u.Quantity[dims.L / dims.T**0.5], "*#batch"] = eqx.field()
    r"""Poincare phi-like variable."""

    z: BBtLength = eqx.field(converter=u.Quantity["length"].from_)
    r"""Height :math:`z \in (-\infty,+\infty)`."""

    dt_rho: ct.BBtSpeed = eqx.field(converter=u.Quantity["speed"].from_)
    r"""Cyindrical radial speed :math:`d\rho/dt \in [-\infty, \infty]."""

    dt_pp_phi: Real[u.Quantity[dims.L / dims.T**0.5], "*#batch"] = eqx.field()
    r"""Poincare phi-like velocity variable."""

    dt_z: ct.BBtSpeed = eqx.field(converter=u.Quantity["speed"].from_)
    r"""Vertical speed :math:`dz/dt \in [-\infty, \infty]."""

    @classmethod
    def _dimensionality(cls) -> int:
        """Dimensionality of the vector.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.vecs.PoincarePolarVector._dimensionality()
        6

        """
        return 6



"""Frames sub-package."""

__all__ = ("frame_transform_op", "frame_of")

from typing import Any

from plum import dispatch


@dispatch.abstract
def frame_transform_op(from_frame: Any, to_frame: Any, /) -> Any:
    """Make a frame transform.

    Parameters
    ----------
    from_frame : AbstractReferenceFrame
        The reference frame to transform from.
    to_frame : AbstractReferenceFrame
        The reference frame to transform to.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv
    >>> import coordinax.frames as cxf

    >>> alice = cxf.Alice()
    >>> bob = cxf.Bob()

    >>> op = cxf.frame_transform_op(alice, bob)
    >>> op
    Pipe(( ... ))

    >>> q = cxv.CartesianPos3D.from_([1, 2, 3], "kpc")
    >>> t = u.Quantity(1, "yr")
    >>> print(op(t, q)[1])
    <CartesianPos3D: (x, y, z) [kpc]
        [1. 2. 3.]>

    """


@dispatch.abstract
def frame_of(obj: Any, /) -> Any:
    """Get the frame of an object."""



"""Space."""

__all__ = ("KinematicSpace",)

from .core import KinematicSpace

# Register by import
# isort: split
from . import (
    register_dataclassish,  # noqa: F401
    register_primitives,  # noqa: F401
    register_vectorapi,  # noqa: F401
)



"""Galilean coordinate transformations."""

__all__ = ["GalileanSpatialTranslation"]


from dataclasses import replace
from typing import Any, Literal, cast, final

import equinox as eqx
import jax
import wadler_lindig as wl
from plum import convert, dispatch

import quaxed.numpy as jnp
import unxt as u

from .base import AbstractGalileanOperator
from coordinax._src.operators.base import AbstractOperator
from coordinax._src.operators.identity import Identity
from coordinax._src.vectors.base_pos import AbstractPos
from coordinax._src.vectors.base_vel import AbstractVel
from coordinax._src.vectors.d1 import CartesianPos1D
from coordinax._src.vectors.d2 import CartesianPos2D
from coordinax._src.vectors.d3 import CartesianPos3D
from coordinax._src.vectors.d4 import FourVector
from coordinax._src.vectors.private_api import spatial_component

##############################################################################
# Spatial Translations


def converter(x: Any) -> AbstractPos:
    """Convert for the spatial translation operator."""
    out: AbstractPos | None
    if isinstance(x, GalileanSpatialTranslation):
        out = x.delta_q
    elif isinstance(x, AbstractPos):
        out = x
    elif isinstance(x, u.AbstractQuantity):
        shape: tuple[int, ...] = x.shape
        match shape:
            case (1,):
                out = cast("AbstractPos", CartesianPos1D.from_(x))
            case (2,):
                out = cast("AbstractPos", CartesianPos2D.from_(x))
            case (3,):
                out = cast("AbstractPos", CartesianPos3D.from_(x))
            case _:
                msg = f"Cannot convert {x} to a spatial translation vector."
                raise TypeError(msg)

    if out is None:
        msg = f"Cannot convert {x} to a spatial translation vector."
        raise TypeError(msg)
    return out


@final
class GalileanSpatialTranslation(AbstractGalileanOperator):
    r"""Operator for Galilean spatial translations.

    The coordinate transform is given by:

    .. math::

        (t,\mathbf{x}) \mapsto (t+s, \mathbf{x} + \mathbf {a})

    Parameters
    ----------
    translation
        The spatial translation vector. This parameters accepts either a
        `vector.AbstractPos3D` instance or uses
        `coordinax.vecs.CartesianPos3D.from_` to enable a variety of more
        convenient input types to create a Cartesian vector. See
        `coordinax.vecs.CartesianPos3D` for details.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    We can then create a spatial translation operator:

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1.0, 2.0, 3.0], "km")
    >>> op
    GalileanSpatialTranslation(CartesianPos3D( ... ))

    Note that the translation is a `coordinax.vecs.CartesianPos3D`, which was
    constructed from a 1D array, using `coordinax.vecs.CartesianPos3D.from_`. We
    can also construct it directly, which allows for other vector types.

    >>> shift = cx.SphericalPos(r=u.Quantity(1.0, "km"),
    ...                         theta=u.Quantity(jnp.pi/2, "rad"),
    ...                         phi=u.Quantity(0, "rad"))
    >>> op = cx.ops.GalileanSpatialTranslation(shift)
    >>> op
    GalileanSpatialTranslation(SphericalPos( ... ))

    Translation operators can be applied to any `vector.AbstractPos`:

    >>> q = cx.CartesianPos3D.from_([0, 0, 0], "km")
    >>> op(q)
    CartesianPos3D( ... )

    Actually, the operator is very flexible and can be applied to many types of
    input. Let's work up the type ladder:

    - `unxt.Quantity`:

    >>> q = u.Quantity([0, 0, 0], "km")
    >>> op(q).value.round(2)
    Array([ 1.,  0., -0.], dtype=float32)

    `coordinax.ops.GalileanSpatialTranslation` can be used for other dimensional
    vectors as well:

    - 1D:

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1], "km")
    >>> q = u.Quantity([0], "km")
    >>> op(q)
    Quantity(Array([1], dtype=int32), unit='km')

    >>> vec = cx.vecs.CartesianPos1D.from_(q).vconvert(cx.vecs.RadialPos)
    >>> op(vec)
    RadialPos(r=Distance(1, unit='km'))

    - 2D:

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1, 2], "km")
    >>> q = u.Quantity([0, 0], "km")
    >>> op(q)
    Quantity(Array([1, 2], dtype=int32), unit='km')

    >>> vec = cx.vecs.CartesianPos2D.from_(q).vconvert(cx.vecs.PolarPos)
    >>> op(vec)
    PolarPos(r=Distance(2.236068, unit='km'), phi=Angle(1.1071488, unit='rad'))

    - 3D:

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1, 2, 3], "km")
    >>> q = u.Quantity([0, 0, 0], "km")
    >>> op(q)
    Quantity(Array([1, 2, 3], dtype=int32), unit='km')

    >>> vec = cx.CartesianPos3D.from_(q).vconvert(cx.SphericalPos)
    >>> op(vec)
    SphericalPos(
      r=Distance(3.7416575, unit='km'),
      theta=Angle(0.64052236, unit='rad'),
      phi=Angle(1.1071488, unit='rad')
    )

    Many operators are time dependent and require a time argument. This operator
    is time independent and will pass through the time argument:

    >>> t = u.Quantity(0, "Gyr")
    >>> op(t, q)[0] is t
    True

    `coordinax.ops.GalileanSpatialTranslation` can be applied to other input
    types. Let's work up the type ladder:

    - `jax.Array`: Note that since the operator is unitful but the

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1, 2, 3], "km")

    >>> q = jnp.array([0, 0, 0])

    """

    delta_q: AbstractPos = eqx.field(converter=converter)
    """The spatial translation.

    This parameters accepts either a `vector.AbstractVector` instance or
    uses a Cartesian vector from_ to enable a variety of more convenient
    input types to create a Cartesian vector. See
    `coordinax.vecs.CartesianPos3D.from_` for an example when doing a 3D
    translation.
    """

    # -------------------------------------------

    @property
    def is_inertial(self) -> Literal[True]:
        """Galilean translation is an inertial frame-preserving transformation.

        Examples
        --------
        >>> import coordinax.ops as cxo

        >>> op = GalileanSpatialTranslation.from_([1, 1, 1], "km")

        >>> op.is_inertial
        True

        """
        return True

    @property
    def inverse(self) -> "GalileanSpatialTranslation":
        """The inverse of the operator.

        Examples
        --------
        >>> import coordinax as cx

        >>> op = cx.ops.GalileanSpatialTranslation.from_([1, 1, 1], "km")

        >>> op.inverse
        GalileanSpatialTranslation(CartesianPos3D( ... ))

        >>> print(op.inverse.delta_q)
        <CartesianPos3D: (x, y, z) [km]
            [-1 -1 -1]>

        """
        return GalileanSpatialTranslation(-self.delta_q)

    # -------------------------------------------
    # Arithmetic operations

    def __neg__(self: "GalileanSpatialTranslation") -> "GalileanSpatialTranslation":
        """Negate the rotation.

        Examples
        --------
        >>> import quaxed.numpy as jnp
        >>> import coordinax as cx

        >>> op = cx.ops.GalileanSpatialTranslation.from_([1, 0, 0], "km")
        >>> print((-op).delta_q)
        <CartesianPos3D: (x, y, z) [km]
            [-1 0 0]>

        """
        return replace(self, delta_q=-self.delta_q)

    # -------------------------------------------

    def __pdoc__(self, **kwargs: Any) -> wl.AbstractDoc:
        """Return the Wadler-Lindig representation."""
        return (
            wl.TextDoc(f"{self.__class__.__name__}(")
            + wl.pdoc(self.delta_q, **kwargs)
            + wl.TextDoc(")")
        )


# ======================================================================
# Call dispatches

# ---------------------------
# Fundamental dispatches


@AbstractOperator.__call__.dispatch
def call(self: GalileanSpatialTranslation, q: AbstractPos, /, **__: Any) -> AbstractPos:
    """Apply the translation to the coordinates.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1, 1, 1], "km")

    >>> q = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> t = u.Quantity(0, "Gyr")
    >>> newq = op(q)
    >>> print(newq)
    <CartesianPos3D: (x, y, z) [km]
        [2 3 4]>

    """
    return cast("AbstractPos", q + self.delta_q)


@AbstractOperator.__call__.dispatch
def call(
    self: GalileanSpatialTranslation,
    t: u.AbstractQuantity,
    q: AbstractPos,
    /,
    **__: Any,
) -> tuple[u.AbstractQuantity, AbstractPos]:
    """Apply the translation to the coordinates.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1, 1, 1], "km")

    >>> q = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> t = u.Quantity(0, "Gyr")
    >>> newt, newq = op(t, q)
    >>> print(newq)
    <CartesianPos3D: (x, y, z) [km]
        [2 3 4]>

    This spatial translation is time independent.

    >>> t = u.Quantity(1, "Gyr")
    >>> op(t, q)[1].x == newq.x
    Array(True, dtype=bool)

    """
    return t, q + self.delta_q


# ---------------------------


@AbstractOperator.__call__.dispatch
def call(self: GalileanSpatialTranslation, v4: FourVector, /, **__: Any) -> AbstractPos:
    """Apply the translation to the coordinates.

    Examples
    --------
    >>> import coordinax as cx

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1, 1, 1], "km")

    >>> v4 = cx.FourVector.from_([0, 1, 2, 3], "km")
    >>> newv4 = op(v4)
    >>> print(newv4)
    <FourVector: (t[s], q=(x, y, z) [km])
        [0. 2. 3. 4.]>

    """
    return replace(v4, q=v4.q + self.delta_q)


@jax.jit
@AbstractOperator.__call__.dispatch
def call(
    self: GalileanSpatialTranslation, qvec: AbstractPos, pvec: AbstractVel, /, **__: Any
) -> tuple[AbstractPos, AbstractVel]:
    r"""Apply the translation to the coordinates.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1, 1, 1], "km")

    >>> q = cx.CartesianPos3D.from_([0, 0, 0], "km")
    >>> p = cx.CartesianVel3D.from_([1, 2, 3], "km/s")
    >>> newq, newp = op(q, p)
    >>> print(newq, newp, sep="\n")
    <CartesianPos3D: (x, y, z) [km]
        [1 1 1]>
    <CartesianVel3D: (x, y, z) [km / s]
        [1. 2. 3.]>

    >>> q = cx.CartesianPos3D.from_([[0, 0, 0], [0, 1, 0]], "km")
    >>> newq, newp = op(q, p)
    >>> print(newq, newp, sep="\n")
    <CartesianPos3D: (x, y, z) [km]
        [[1 1 1]
         [1 2 1]]>
    <CartesianVel3D: (x, y, z) [km / s]
        [[1. 2. 3.]
         [1. 2. 3.]]>

    """
    newqvec = self(qvec)

    # TODO: figure out how to do this in general, then all these dispatches
    # can be consolidated. And do it on vectors, not the quantities.
    #
    # Translate the velocity (this operator will have no effect on the
    # velocity).
    # 1. convert to a Quantity in Cartesian coordinates.
    qvec = spatial_component(qvec)
    pvec = eqx.error_if(
        pvec,
        pvec._dimensionality() != qvec._dimensionality(),
        "The position and velocity vectors must have the same dimensionality.",
    )

    qcart_cls = qvec.cartesian_type
    pcart_cls = pvec.cartesian_type
    q = convert(qvec.vconvert(qcart_cls), u.Quantity)
    p = convert(pvec.vconvert(pcart_cls, qvec), u.Quantity)
    # 1.5 flatten all but the last axis  # TODO: not need to flatten
    batch = jnp.broadcast_shapes(q.shape[:-1], p.shape[:-1])
    q, p = jnp.reshape(q, (-1, q.shape[-1])), jnp.reshape(p, (-1, q.shape[-1]))
    # 1.5 cast to float dtype  # TODO: more careful casting
    q, p = q.astype(float, copy=False), p.astype(float, copy=False)
    # 2. create the Jacobian of the operation on the position
    jac = jax.vmap(u.experimental.jacfwd(self.__call__, argnums=0, units=(q.unit,)))(q)
    # 3. apply the Jacobian to the velocity
    newp = jnp.einsum("bmn,bn->bm", jac, p)
    newp = jnp.reshape(newp, (*batch, newp.shape[-1]))
    # 4. convert the Quantity back to a Cartesian vector
    newpvec = pcart_cls.from_(newp)
    # 5. convert the Quantity to the original vector type
    newpvec = newpvec.vconvert(type(pvec), newqvec)

    return newqvec, newpvec


@AbstractOperator.__call__.dispatch
def call(
    self: GalileanSpatialTranslation,
    q: u.AbstractQuantity,
    p: u.Quantity["speed"],
    /,
    **__: Any,
) -> tuple[u.AbstractQuantity, u.AbstractQuantity]:
    r"""Apply the translation to the coordinates.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1, 1, 1], "km")

    >>> q = cx.CartesianPos3D.from_([0, 0, 0], "km")
    >>> p = cx.CartesianVel3D.from_([1, 2, 3], "km/s")
    >>> newq, newp = op(q, p)
    >>> print(newq, newp, sep="\n")
    <CartesianPos3D: (x, y, z) [km]
        [1 1 1]>
    <CartesianVel3D: (x, y, z) [km / s]
        [1. 2. 3.]>

    """
    newq = self(q)

    # TODO: figure out how to do this in general, then all these dispatches
    # can be consolidated. And do it on vectors, not the quantities.
    #
    # Translate the velocity (this operator will have no effect on the
    # velocity).
    # 2. create the Jacobian of the operation on the position
    jac = u.experimental.jacfwd(self.__call__, argnums=0, units=(q.unit,))(q)
    # 3. apply the Jacobian to the velocity
    newp = jac @ p

    return newq, newp


# ======================================================================
# Simplification


@dispatch
def simplify_op(
    op: GalileanSpatialTranslation, /, **kwargs: Any
) -> GalileanSpatialTranslation | Identity:
    """Simplify a Galilean spatial translation operator.

    Examples
    --------
    >>> import coordinax as cx

    An operator with real effect cannot be simplified:

    >>> op = cx.ops.GalileanSpatialTranslation.from_([1, 0, 0], "m")
    >>> cx.ops.simplify_op(op)
    GalileanSpatialTranslation(CartesianPos3D( ... ))

    An operator with no effect can be simplified:

    >>> op = cx.ops.GalileanSpatialTranslation.from_([0, 0, 0], "m")
    >>> cx.ops.simplify_op(op)
    Identity()

    """
    # Check if the translation is zero.
    if jnp.allclose(convert(op.delta_q, u.Quantity).value, 0, **kwargs):
        return Identity()
    return op


@dispatch
def simplify_op(
    op1: GalileanSpatialTranslation, op2: GalileanSpatialTranslation, /
) -> GalileanSpatialTranslation:
    """Combine two spatial translations into a single translation.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> op1 = cx.ops.GalileanSpatialTranslation.from_([1, 0, 0], "km")
    >>> op2 = cx.ops.GalileanSpatialTranslation.from_([0, 1, 0], "km")

    >>> op3 = cx.ops.simplify_op(op1, op2)
    >>> op3
    GalileanSpatialTranslation(CartesianPos3D( ... ))

    >>> op3.delta_q == op1.delta_q + op2.delta_q
    Array(True, dtype=bool)

    """
    return GalileanSpatialTranslation(op1.delta_q + op2.delta_q)



r"""Reference frames and transformations between them.

Examples
--------
>>> import quaxed.numpy as jnp
>>> import unxt as u
>>> import coordinax as cx
>>> import coordinax.frames as cxf

>>> frame1 = cxf.Alice()
>>> frame2 = cxf.Bob()

Let's transform a position from Alice's frame to Bob's frame:

>>> op = cxf.frame_transform_op(frame1, frame2)
>>> op
Pipe((
    GalileanSpatialTranslation(CartesianPos3D( ... )),
    GalileanBoost(CartesianVel3D( ... ))
))

>>> q_alice = cx.CartesianPos3D.from_([0, 0, 0], "km")
>>> t = u.Quantity(2.5, "yr")
>>> _, q_bob = op(t, q_alice)
>>> print(q_bob)
<CartesianPos3D: (x, y, z) [km]
    [2.129e+13 1.000e+04 0.000e+00]>

Now let's create a new transformed frame and work with it:

>>> R = cx.ops.GalileanRotation([[0., -1, 0], [1, 0, 0], [0, 0, 1]])
>>> frame = cxf.TransformedReferenceFrame(frame1, R)
>>> frame
TransformedReferenceFrame(
    base_frame=Alice(), xop=GalileanRotation(rotation=f32[3,3])
)

Let's transform a position from the base frame to the transformed frame:

>>> op = cxf.frame_transform_op(cxf.Alice(), frame)

>>> q_icrs = cx.CartesianPos3D.from_([1, 0, 0], "kpc")
>>> q_frame = op(q_icrs)
>>> print(q_frame)
<CartesianPos3D: (x, y, z) [kpc]
    [ 0. -1.  0.]>

>>> op.inverse(q_frame) == q_icrs
Array(True, dtype=bool)

This can also transform a velocity:

>>> v_icrs = cx.CartesianVel3D.from_([1, 0, 0], "km/s")
>>> q_frame, v_frame = op(q_icrs, v_icrs)
>>> print(q_frame, v_frame, sep="\n")
<CartesianPos3D: (x, y, z) [kpc]
    [ 0. -1.  0.]>
<CartesianVel3D: (x, y, z) [km / s]
    [ 0. -1.  0.]>

>>> op.inverse(q_frame, v_frame) == (q_icrs, v_icrs)
True

"""

from .setup_package import install_import_hook

with install_import_hook("coordinax.frames"):
    from ._src.frames import (
        AbstractCoordinate,
        AbstractReferenceFrame,
        Alice,
        Bob,
        Coordinate,
        FrameTransformError,
        FriendOfAlice,
        NoFrame,
        TransformedReferenceFrame,
        frame_of,
        frame_transform_op,
    )

    # Frames from external packages
    # isort: split
    from . import _coordinax_space_frames
    from ._coordinax_space_frames import *  # noqa: F403


# Defined here b/c it's mutated by `_coordinax_space_frames`
__all__ = [
    # Frames
    "AbstractReferenceFrame",
    "FrameTransformError",
    "NoFrame",
    "TransformedReferenceFrame",
    "frame_transform_op",
    "frame_of",
    # Example frames
    "Alice",
    "FriendOfAlice",
    "Bob",
    # Coordinates
    "AbstractCoordinate",
    "Coordinate",
]
__all__ += _coordinax_space_frames.__all__

# clean up namespace
del _coordinax_space_frames, install_import_hook



"""`coordinax.distance` module."""

__all__ = (
    # Classes
    "AbstractDistance",
    "Distance",
    "DistanceModulus",
    "Parallax",
    # funcs
    "distance",
    "parallax",
    "distance_modulus",
)

from .setup_package import install_import_hook

with install_import_hook("coordinax.distance"):
    from ._src.distances import (
        AbstractDistance,
        Distance,
        DistanceModulus,
        Parallax,
        distance,
        distance_modulus,
        parallax,
    )


del install_import_hook



"""Representation of coordinates in different systems."""

__all__ = ("AbstractPos4D",)


from coordinax._src.vectors.base_pos import AbstractPos


class AbstractPos4D(AbstractPos):
    """Abstract representation of 4D coordinates in different systems."""

    @classmethod
    def _dimensionality(cls) -> int:
        """Dimensionality of the vector.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.vecs.AbstractPos4D._dimensionality()
        4

        """
        return 4



"""Transformations between representations."""

__all__: tuple[str, ...] = ()

import functools as ft
from typing import Any

import equinox as eqx
import jax
import jax.tree as jtu
from jaxtyping import Array, Float, Real
from plum import dispatch

import quaxed.numpy as jnp
import unxt as u
from unxt.quantity import BareQuantity, is_any_quantity

import coordinax._src.vectors.custom_types as ct
from coordinax._src.vectors import api, d1, d2, d3
from coordinax._src.vectors.base import AbstractVector
from coordinax._src.vectors.base_acc import AbstractAcc
from coordinax._src.vectors.base_pos import AbstractPos
from coordinax._src.vectors.base_vel import AbstractVel

pos_jac_fn = jax.vmap(
    jax.jacfwd(  # ↓ aux as positional argument
        lambda to_cls, from_cls, p, in_aux, out_aux: api.vconvert(
            to_cls, from_cls, p, in_aux=in_aux, out_aux=out_aux
        ),
        argnums=2,
        has_aux=True,
    ),
    in_axes=(None, None, 0, None, None),
)

is_q_or_arr = lambda x: is_any_quantity(x) or eqx.is_array(x)  # noqa: E731


def compute_jac(
    to_pos_cls: type[AbstractPos],
    from_pos_cls: type[AbstractPos],
    p_pos: ct.ParamsDict,
    /,
    *,
    in_aux: ct.AuxDict,
    out_aux: ct.AuxDict,
) -> dict[str, dict[str, u.AbstractQuantity]]:
    """Compute the Jacobian of the transformation."""
    # Compute the Jacobian of the transformation.
    # NOTE: this is using Quantities. Using raw arrays is ~20x faster.
    jac, _ = pos_jac_fn(to_pos_cls, from_pos_cls, p_pos, in_aux, out_aux)
    # Restructure the Jacobian:
    # from: ``{to_k: Quantity({from_k: Quantity(dto/dfrom, u_from)}, u_to)}``
    # to  : ``{to_k: {from_k: Quantity(dto/dfrom, u_to/u_from)}}``.
    jac = {
        out_k: {
            k: BareQuantity(v.value, out_v.unit / v.unit)
            for k, v in out_v.value.items()
        }
        for out_k, out_v in jac.items()
    }
    return jac  # noqa: RET504


@ft.partial(jax.jit, inline=True)
def inner_dot(
    inner: dict[str, u.AbstractQuantity],
    vec: dict[str, u.AbstractQuantity],
) -> u.AbstractQuantity:
    """Dot product of two dicts.

    This is a helper function for the `dot_jac_vec` function.

    Parameters
    ----------
    inner
        The first dict.
        The structure is ``{from_k: Quantity(v, unit)}``.
    vec
        The second dict.
        The structure is ``{from_k: Quantity(v, unit)}``.

    """
    return jtu.reduce(
        jnp.add,
        jtu.map(jnp.multiply, inner, vec, is_leaf=is_q_or_arr),
        is_leaf=is_q_or_arr,
    )


@jax.jit
def dot_jac_vec(
    jac: dict[str, dict[str, u.AbstractQuantity]], vec: dict[str, u.AbstractQuantity]
) -> dict[str, u.AbstractQuantity]:
    """Dot product of a Jacobian dict and a vector dict.

    This is a helper function for the `vconvert` function.

    Parameters
    ----------
    jac
        The Jacobian of the transformation.
        The structure is ``{to_k: {from_k: Quantity(v, unit)}}``.
    vec
        The vector to transform.
        The structure is ``{from_k: Quantity(v, unit)}``.

    Examples
    --------
    >>> import unxt as u

    >>> J = {"r": {"x": u.Quantity(1, ""), "y": u.Quantity(2, "")},
    ...      "phi": {"x": u.Quantity(3, "rad/km"), "y": u.Quantity(4, "rad/km")}}

    >>> v = {"x": u.Quantity(1, "km"), "y": u.Quantity(2, "km")}

    >>> dot_jac_vec(J, v)
    {'phi': Quantity(Array(11, dtype=int32, ...), unit='rad'),
     'r': Quantity(Array(5, dtype=int32, ...), unit='km')}

    """
    # TODO: rewrite this by separating the units and the values
    return {k: inner_dot(inner, vec) for k, inner in jac.items()}


@ft.partial(jax.jit, inline=True)
def atleast_1d_float(x: Real[Array, "..."]) -> Float[Array, "..."]:
    """Convert to 1D float array."""
    return jnp.atleast_1d(jnp.astype(x, float, copy=False))


# ============================================================================


@dispatch.multi(
    (type[AbstractVel], type[AbstractVel], ct.ParamsDict, ct.ParamsDict),
    (type[AbstractAcc], type[AbstractAcc], ct.ParamsDict, ct.ParamsDict),
)
def vconvert(
    to_dif_cls: type[AbstractVector],
    from_dif_cls: type[AbstractVector],
    p_dif: ct.ParamsDict,
    p_pos: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
) -> tuple[ct.ParamsDict, ct.OptAuxDict]:
    """AbstractVel1D -> AbstractVel1D.

    Parameters
    ----------
    to_dif_cls
        The target type of the vector differential.
    from_dif_cls
        The type of the vector differential to transform.
    p_dif
        The data of the vector differential to transform.
    p_pos
        The data of the position vector used to transform the differential.
    in_aux
        The input auxiliary data to the transform.
    out_aux
        THe output auxiliary data to the transform.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    Let's start in 1D:

    >>> q = {"x": u.Quantity([1.0], "km")}
    >>> p = {"x": u.Quantity([1.0], "km/s")}
    >>> newp = cxv.vconvert(cxv.RadialVel, cxv.CartesianVel1D, p, q)
    >>> print(newp)
    ({'r': Quantity(Array([1.], dtype=float32), unit='km / s')}, {})

    >>> q = {"x": u.Quantity([1.0], "km")}
    >>> a = {"x": u.Quantity([1.0], "km/s2")}
    >>> newa = cxv.vconvert(cxv.RadialAcc, cxv.CartesianAcc1D, a, q)
    >>> print(newa)
    ({'r': Quantity(Array([1.], dtype=float32), unit='km / s2')}, {})

    """
    # Check the dimensionality
    to_dif_cls = eqx.error_if(
        to_dif_cls,
        from_dif_cls._dimensionality() != to_dif_cls._dimensionality(),
        f"Dimensionality mismatch: cannot convert from {from_dif_cls.__name__} "
        f"(dim={from_dif_cls._dimensionality()}) "
        f"to {to_dif_cls.__name__} (dim={to_dif_cls._dimensionality()}) "
        "as their dimensionalities do not match.",
    )

    # Parse the auxiliary data
    in_aux_ = in_aux or {}
    out_aux_ = out_aux or {}

    # Compute the Jacobian of the position transformation.
    # The position is assumed to be in the type required by the differential to
    # construct the Jacobian. E.g. for CartesianVel1D -> RadialVel, we need the
    # Jacobian of the CartesianPos1D -> RadialPos transform.
    n = -2 if issubclass(to_dif_cls, AbstractAcc) else -1
    to_pos_cls = to_dif_cls.time_nth_derivative_cls(n=n)
    from_pos_cls = from_dif_cls.time_nth_derivative_cls(n=n)
    p_pos = jtu.map(atleast_1d_float, p_pos)
    jac = compute_jac(to_pos_cls, from_pos_cls, p_pos, in_aux=in_aux_, out_aux=out_aux_)

    # Transform the differential
    to_p_dif = dot_jac_vec(jac, p_dif)

    # Reshape the output to the shape of the input
    shape = jnp.broadcast_shapes(*[v.shape for v in p_dif.values()])
    to_p_dif = jtu.map(lambda x: jnp.reshape(x, shape), to_p_dif)

    return to_p_dif, out_aux_


# ============================================================================


# TODO: implement for cross-representations
@dispatch.multi(
    # Velocities
    (type[d1.AbstractVel1D], d1.AbstractVel1D, AbstractPos),
    (type[d2.AbstractVel2D], d2.AbstractVel2D, AbstractPos),
    (type[d3.AbstractVel3D], d3.AbstractVel3D, AbstractPos),
    # Accelerations
    (type[d1.AbstractAcc1D], d1.AbstractAcc1D, AbstractPos),
    (type[d2.AbstractAcc2D], d2.AbstractAcc2D, AbstractPos),
    (type[d3.AbstractAcc3D], d3.AbstractAcc3D, AbstractPos),
)
def vconvert(
    to_dif_cls: type[AbstractVector],
    from_dif: AbstractVector,
    from_pos: AbstractPos,
    /,
    **kwargs: Any,
) -> AbstractVector:
    """Differential -> Differential.

    This is the base case for the transformation of vector differentials.

    Parameters
    ----------
    to_dif_cls
        The target type of the vector differential.
    from_dif
        The vector differential to transform.
    from_pos
        The position vector used to transform the differential.
    **kwargs
        Additional keyword arguments.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    Let's start in 1D:

    >>> q = cxv.CartesianPos1D.from_(1.0, "km")
    >>> p = cxv.CartesianVel1D.from_(1.0, "km/s")
    >>> newp = cxv.vconvert(cxv.RadialVel, p, q)
    >>> print(newp)
    <RadialVel: (r) [km / s]
        [1.]>

    >>> a = cxv.CartesianAcc1D(x=u.Quantity(1.0, "km/s2"))
    >>> print(cxv.vconvert(cxv.RadialAcc, a, p, q))
    <RadialAcc: (r) [km / s2]
        [1.]>

    Now in 2D:

    >>> q = cxv.CartesianPos2D.from_([1.0, 2.0], "km")
    >>> p = cxv.CartesianVel2D.from_([1.0, 2.0], "km/s")
    >>> newp = cxv.vconvert(cxv.PolarVel, p, q)
    >>> print(newp)
    <PolarVel: (r[km / s], phi[rad / s])
        [2.236 0.   ]>

    >>> a = cxv.CartesianAcc2D.from_([1.0, 2.0], "km/s2")
    >>> print(cxv.vconvert(cxv.PolarAcc, a, p, q))
    <PolarAcc: (r[km / s2], phi[rad / s2])
        [2.236 0.   ]>

    And in 3D:

    >>> q = cxv.CartesianPos3D.from_([1.0, 2.0, 3.0], "km")
    >>> p = cxv.CartesianVel3D.from_([1.0, 2.0, 3.0], "km/s")
    >>> newp = cxv.vconvert(cxv.SphericalVel, p, q)
    >>> print(newp.round(2))
    <SphericalVel: (r[km / s], theta[rad / s], phi[rad / s])
        [ 3.74 -0.    0.  ]>

    >>> a = cxv.CartesianAcc3D.from_([1.0, 2.0, 3.0], "km/s2")
    >>> print(cxv.vconvert(cxv.SphericalAcc, a, p, q).round(2))
    <SphericalAcc: (r[km / s2], theta[rad / s2], phi[rad / s2])
        [ 3.74 -0.    0.  ]>

    """
    # Transform the position to the type required by the differential to
    # construct the Jacobian. E.g. if we are transforming CartesianVel1D ->
    # RadialVel, we need the Jacobian of the CartesianPos1D -> RadialPos so the
    # position must be transformed to CartesianPos1D.
    n = -2 if issubclass(to_dif_cls, AbstractAcc) else -1
    in_pos_cls = from_dif.time_nth_derivative_cls(n=n)
    from_posv = vconvert(in_pos_cls, from_pos, **kwargs)

    # The position dictionary contains both the fields and the auxiliary
    # data. We need to separate them.
    pos_d = from_posv.asdict()
    in_aux = {k: pos_d.pop(k) for k in from_posv._AUX_FIELDS}

    # The output auxiliary data must either be provided through the kwargs, or
    # if the input position is the correct type, we can recover it from there.
    # For example:
    # >>> prolatespheroidalvel.vconvert(cxv.CartesianVel3D,
    # prolatespheroidalpos) Then prolatespheroidalpos has the required auxiliary
    # data. We prefer to use the kwargs, as this was used to convert from_pos to
    # from_posv.
    to_pos_cls = to_dif_cls.time_nth_derivative_cls(n=n)
    out_aux = (
        {k: getattr(from_pos, k) for k in to_pos_cls._AUX_FIELDS}
        if to_pos_cls is type(from_pos)
        else {}
    )
    out_aux |= {k: v for k, v in kwargs.items() if k in to_pos_cls._AUX_FIELDS}

    # Perform the transformation
    # TODO: add  df(q)/dt, which is 0 for all current transforms
    to_dif_p, to_dif_aux = vconvert(
        to_dif_cls,
        type(from_dif),
        from_dif.asdict(),
        pos_d,
        in_aux=in_aux,
        out_aux=out_aux,
    )

    # Reconstruct the vector
    to_dif = to_dif_cls(
        **to_dif_p,
        **{k: v for k, v in to_dif_aux.items() if k in to_dif_cls._AUX_FIELDS},
    )
    to_dif = to_dif.reshape(from_dif.shape)  # reshape to original shape

    return to_dif  # noqa: RET504


# ===================================================================


# TODO: implement for cross-representations
@dispatch.multi(
    # N-D -> N-D
    (type[d1.AbstractAcc1D], d1.AbstractAcc1D, AbstractVel, AbstractPos),
    (type[d2.AbstractAcc2D], d2.AbstractAcc2D, AbstractVel, AbstractPos),
    (type[d3.AbstractAcc3D], d3.AbstractAcc3D, AbstractVel, AbstractPos),
)
def vconvert(
    to_acc_cls: type[AbstractAcc],
    from_acc: AbstractAcc,
    from_vel: AbstractVel,
    from_pos: AbstractPos,
    /,
    **kwargs: Any,
) -> AbstractAcc:
    r"""AbstractAcc -> Cartesian -> AbstractAcc.

    This is the base case for the transformation of accelerations.

    Let $\mathbf{x}$ be a position vector in one representation $\mathbf{y}$ a
    position vector in another representation related by:

    $$ \mathbf{y} = f(\mathbf{x}) $$

    where $f$ is a differentiable function mapping between the coordinate
    systems.

    The Jacobian matrix $J$ of the transformation is:

    $$ J = \frac{\partial \mathbf{y}}{\partial \mathbf{x}}$$

    The coordinate transformation of the acceleration is given by the chain
    rule:

    $$ \ddot{\mathbf{y}} = \dot{J} \dot{\mathbf{x}} + J \ddot{\mathbf{x}}$$

    where $\dot{J}$ is the time derivative of the Jacobian matrix. This function
    assumes that the representation conversion is time-invariant, so $\dot{J} =
    0$. Thus, the transformation simplifies to:

    $$ \ddot{\mathbf{y}} = J \ddot{\mathbf{x}}$$

    This function implements this transformation.

    Parameters
    ----------
    to_acc_cls
        The target type of the vector acceleration.
    from_acc
        The vector acceleration to transform.
    from_vel
        The velocity vector used to transform the acceleration.
    from_pos
        The position vector used to transform the acceleration.
    **kwargs
        Additional keyword arguments.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    Let's start in 1D:

    >>> q = cxv.CartesianPos1D(x=u.Quantity(1.0, "km"))
    >>> p = cxv.CartesianVel1D(x=u.Quantity(1.0, "km/s"))
    >>> a = cxv.CartesianAcc1D(x=u.Quantity(1.0, "km/s2"))
    >>> print(cxv.vconvert(cxv.RadialAcc, a, p, q))
    <RadialAcc: (r) [km / s2]
        [1.]>

    Now in 2D:

    >>> q = cxv.CartesianPos2D.from_([1.0, 2.0], "km")
    >>> p = cxv.CartesianVel2D.from_([1.0, 2.0], "km/s")
    >>> a = cxv.CartesianAcc2D.from_([1.0, 2.0], "km/s2")
    >>> print(cxv.vconvert(cxv.PolarAcc, a, p, q))
    <PolarAcc: (r[km / s2], phi[rad / s2])
        [2.236 0.   ]>

    And in 3D:

    >>> q = cxv.CartesianPos3D.from_([1.0, 2.0, 3.0], "km")
    >>> p = cxv.CartesianVel3D.from_([1.0, 2.0, 3.0], "km/s")
    >>> a = cxv.CartesianAcc3D.from_([1.0, 2.0, 3.0], "km/s2")
    >>> print(cxv.vconvert(cxv.SphericalAcc, a, p, q).round(2))
    <SphericalAcc: (r[km / s2], theta[rad / s2], phi[rad / s2])
        [ 3.74 -0.    0.  ]>

    """
    del from_vel  # unused
    return vconvert(to_acc_cls, from_acc, from_pos, **kwargs)



"""Type hints for `coordinax.distance`."""

__all__: tuple[str, ...] = ()


from jaxtyping import Shaped

import unxt as u

from .base import AbstractDistance

BBtLength = Shaped[u.Quantity["length"], "*#batch"]
BatchableDistance = Shaped[AbstractDistance, "*#batch"]



"""Representation of coordinates in different systems."""

__all__: tuple[str, ...] = ()

from typing import TypeAlias

from astropy.units import (
    CompositeUnit as AstropyCompositeUnit,
    Unit as AstropyUnit,
    UnitBase as AstropyUnitBase,
)
from jaxtyping import Real, Shaped

import unxt as u

Shape: TypeAlias = tuple[int, ...]
Unit: TypeAlias = AstropyUnit | AstropyUnitBase | AstropyCompositeUnit

BBtScalarQ = Shaped[u.AbstractQuantity, "*#batch"]

ScalarTime = Real[u.Quantity["time"], ""]

BBtTime = Real[u.Quantity["time"], "*#batch"]

BBtArea = Real[u.Quantity["area"], "*#batch"]
BBtKinematicFlux = Real[u.Quantity["diffusivity"], "*#batch"]
BBtSpecificEnergy = Real[u.Quantity["specific energy"], "*#batch"]

BBtLength = Real[u.Quantity["length"], "*#batch"]

BBtSpeed = Real[u.Quantity["speed"], "*#batch"]
BBtAngularSpeed = Real[u.Quantity["angular speed"], "*#batch"]

BBtAcc = Real[u.Quantity["acceleration"], "*#batch"]
BBtAngularAcc = Real[u.Quantity["angular acceleration"], "*#batch"]


TimeBatchOrScalar = ScalarTime | BBtTime



"""Transformations for Vectors."""

__all__: tuple[str, ...] = ()

from typing import Any, cast
from warnings import warn

import equinox as eqx
from plum import dispatch

import coordinax._src.vectors.custom_types as ct
from coordinax._src.vectors import d1, d2, d3, d4, dn
from coordinax._src.vectors.api import vconvert
from coordinax._src.vectors.base import AbstractVector
from coordinax._src.vectors.base_acc import AbstractAcc
from coordinax._src.vectors.base_pos import AbstractPos
from coordinax._src.vectors.base_vel import AbstractVel
from coordinax._src.vectors.exceptions import IrreversibleDimensionChange
from coordinax._src.vectors.private_api import combine_aux, wrap_vconvert_impl_params


def get_params_and_aux(obj: AbstractVector, /) -> tuple[ct.ParamsDict, ct.AuxDict]:
    # Get the parameters and auxiliary data
    p_and_aux = cast("dict[str, Any]", obj.asdict())
    # Separate the parameters from the auxiliary data
    comps = obj.components
    p = {k: p_and_aux.pop(k) for k in tuple(p_and_aux) if k in comps}
    in_aux = p_and_aux  # popped all the params out
    return p, in_aux


# =============================================================================


@dispatch.multi(
    (type[AbstractPos], AbstractPos),
    (type[AbstractVel], AbstractVel),
    (type[AbstractAcc], AbstractAcc),
)
def vconvert(
    target: type[AbstractVector],
    current: AbstractVector,
    /,
    units: ct.OptUSys = None,
    **out_aux: Any,
) -> AbstractVector:
    """AbstractPos -> vconvert_impl -> AbstractPos.

    This is the base case for the transformation of position vectors.

    """
    # Get the parameters and auxiliary data
    p_and_aux = cast("dict[str, Any]", current.asdict())
    # Separate the parameters from the auxiliary data
    comps = current.components
    p = {k: p_and_aux.pop(k) for k in tuple(p_and_aux) if k in comps}
    in_aux = p_and_aux  # popped all the params out

    # Parameters can be passed by kwarg, so we need to filter them out
    # from the auxiliary data. E.g. for RadialPos -> CartesianPos2D
    # we need to specify `y` by kwarg. It is a param, not an aux.
    p_by_kw = {k: out_aux.pop(k) for k in tuple(out_aux) if k in target.components}
    p_by_kw = eqx.error_if(p_by_kw, any(set(p_by_kw) & set(p)), "overlap in params.")
    p = p | p_by_kw

    # Convert the parameters, using the auxiliary data
    p, aux = vconvert(
        target, type(current), p, in_aux=in_aux, out_aux=out_aux, units=units
    )

    # Build the new vector
    return target(**(aux or {}), **p)


@dispatch.multi(
    (type[AbstractPos], type[AbstractPos], AbstractPos),
    (type[AbstractVel], type[AbstractVel], AbstractVel),
    (type[AbstractAcc], type[AbstractAcc], AbstractAcc),
)
def vconvert(
    to_vector: type[AbstractVector],
    from_vector: type[AbstractVector],
    current: AbstractVector,
    /,
    **out_aux: Any,
) -> AbstractVector:
    """Convert from one position vector to another.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> x = cxv.CartesianPos2D.from_([3, 4], "km")
    >>> y = cxv.vconvert(cxv.PolarPos, cxv.CartesianPos2D, x)
    >>> print(y)
    <PolarPos: (r[km], phi[rad])
        [5.    0.927]>

    """
    current = eqx.error_if(
        current,
        not isinstance(current, from_vector),
        f"{from_vector} expected, got {type(current)}.",
    )
    return vconvert(to_vector, current, **out_aux)


# =============================================================================
# Self transform


@dispatch.multi(
    # Positions
    (type[d1.CartesianPos1D], type[d1.CartesianPos1D], ct.ParamsDict),
    (type[d1.RadialPos], type[d1.RadialPos], ct.ParamsDict),
    (type[d2.CartesianPos2D], type[d2.CartesianPos2D], ct.ParamsDict),
    (type[d2.PolarPos], type[d2.PolarPos], ct.ParamsDict),
    (type[d2.TwoSpherePos], type[d2.TwoSpherePos], ct.ParamsDict),
    (type[d3.CartesianPos3D], type[d3.CartesianPos3D], ct.ParamsDict),
    (type[d3.CylindricalPos], type[d3.CylindricalPos], ct.ParamsDict),
    (type[d3.SphericalPos], type[d3.SphericalPos], ct.ParamsDict),
    (type[d3.LonLatSphericalPos], type[d3.LonLatSphericalPos], ct.ParamsDict),
    (type[d3.MathSphericalPos], type[d3.MathSphericalPos], ct.ParamsDict),
    (type[dn.PoincarePolarVector], type[dn.PoincarePolarVector], ct.ParamsDict),
    (type[dn.CartesianPosND], type[dn.CartesianPosND], ct.ParamsDict),
    # Velocities
    (type[d1.CartesianVel1D], type[d1.CartesianVel1D], ct.ParamsDict),
    (type[d1.RadialVel], type[d1.RadialVel], ct.ParamsDict),
    (type[d2.CartesianVel2D], type[d2.CartesianVel2D], ct.ParamsDict),
    (type[d2.PolarVel], type[d2.PolarVel], ct.ParamsDict),
    (type[d2.TwoSphereVel], type[d2.TwoSphereVel], ct.ParamsDict),
    (type[d3.CartesianVel3D], type[d3.CartesianVel3D], ct.ParamsDict),
    (type[d3.CylindricalVel], type[d3.CylindricalVel], ct.ParamsDict),
    (type[d3.SphericalVel], type[d3.SphericalVel], ct.ParamsDict),
    (type[d3.LonLatSphericalVel], type[d3.LonLatSphericalVel], ct.ParamsDict),
    (type[d3.LonCosLatSphericalVel], type[d3.LonCosLatSphericalVel], ct.ParamsDict),
    (type[d3.MathSphericalVel], type[d3.MathSphericalVel], ct.ParamsDict),
    (type[dn.CartesianVelND], type[dn.CartesianVelND], ct.ParamsDict),
    # Accelerations
    (type[d1.CartesianAcc1D], type[d1.CartesianAcc1D], ct.ParamsDict),
    (type[d1.RadialAcc], type[d1.RadialAcc], ct.ParamsDict),
    (type[d2.CartesianAcc2D], type[d2.CartesianAcc2D], ct.ParamsDict),
    (type[d3.CartesianAcc3D], type[d3.CartesianAcc3D], ct.ParamsDict),
    (type[d3.CylindricalAcc], type[d3.CylindricalAcc], ct.ParamsDict),
    (type[d3.SphericalAcc], type[d3.SphericalAcc], ct.ParamsDict),
    (type[d3.LonLatSphericalAcc], type[d3.LonLatSphericalAcc], ct.ParamsDict),
    (type[d3.MathSphericalAcc], type[d3.MathSphericalAcc], ct.ParamsDict),
    (type[dn.CartesianAccND], type[dn.CartesianAccND], ct.ParamsDict),
)
def vconvert(
    to_vector: type[AbstractVector],
    from_vector: type[AbstractVector],
    params: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """Self transform.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    ## 1D:

    >>> params = {"x": 1}
    >>> cxv.vconvert(cxv.CartesianPos1D, cxv.CartesianPos1D, params)
    ({'x': 1}, {})

    >>> params = {"r": 1}
    >>> cxv.vconvert(cxv.RadialPos, cxv.RadialPos, params)
    ({'r': 1}, {})

    ## 2D:

    >>> params = {"x": 1, "y": 2}
    >>> cxv.vconvert(cxv.CartesianPos2D, cxv.CartesianPos2D, params)
    ({'x': 1, 'y': 2}, {})

    >>> params = {"r": 1, "phi": 10}
    >>> cxv.vconvert(cxv.PolarPos, cxv.PolarPos, params)
    ({'r': 1, 'phi': 10}, {})

    >>> params = {"theta": 10, "phi": 14}
    >>> cxv.vconvert(cxv.TwoSpherePos, cxv.TwoSpherePos, params)
    ({'theta': 10, 'phi': 14}, {})

    ## 3D:

    >>> params = {"x": 1, "y": 2, "z": 3}
    >>> cxv.vconvert(cxv.CartesianPos3D, cxv.CartesianPos3D, params)
    ({'x': 1, 'y': 2, 'z': 3}, {})

    >>> params = {"rho": 1, "phi": 2, "z": 3}
    >>> cxv.vconvert(cxv.CylindricalPos, cxv.CylindricalPos, params)
    ({'rho': 1, 'phi': 2, 'z': 3}, {})

    >>> params = {"r": 1, "theta": 2, "phi": 3}
    >>> cxv.vconvert(cxv.SphericalPos, cxv.SphericalPos, params)
    ({'r': 1, 'theta': 2, 'phi': 3}, {})

    >>> params = {"lon": 1, "lat": 2, "distance": 3}
    >>> cxv.vconvert(cxv.LonLatSphericalPos, cxv.LonLatSphericalPos, params)
    ({'lon': 1, 'lat': 2, 'distance': 3}, {})

    >>> params = {"r": 1, "theta": 2, "phi": 3}
    >>> cxv.vconvert(cxv.MathSphericalPos, cxv.MathSphericalPos, params)
    ({'r': 1, 'theta': 2, 'phi': 3}, {})

    >>> params = {"rho": 1, "pp_phi": 2, "z": 3, "dt_rho": 4, "dt_pp_phi": 5, "dt_z": 6}
    >>> cxv.vconvert(cxv.PoincarePolarVector, cxv.PoincarePolarVector, params)
    ({'rho': 1, 'pp_phi': 2, 'z': 3, 'dt_rho': 4, 'dt_pp_phi': 5, 'dt_z': 6}, {})

    >>> params = {"q": jnp.array([1, 2, 3, 4])}
    >>> cxv.vconvert(cxv.CartesianPosND, cxv.CartesianPosND, params)
    ({'q': Array([1, 2, 3, 4], dtype=int32)}, {})

    """
    return params, combine_aux(in_aux, out_aux)


@dispatch.multi(  # TODO: is the precedence needed?
    # Positions
    (type[d1.CartesianPos1D], d1.CartesianPos1D),
    (type[d1.RadialPos], d1.RadialPos),
    (type[d2.CartesianPos2D], d2.CartesianPos2D),
    (type[d2.PolarPos], d2.PolarPos),
    (type[d2.TwoSpherePos], d2.TwoSpherePos),
    (type[d3.CartesianPos3D], d3.CartesianPos3D),
    (type[d3.CylindricalPos], d3.CylindricalPos),
    (type[d3.SphericalPos], d3.SphericalPos),
    (type[d3.LonLatSphericalPos], d3.LonLatSphericalPos),
    (type[d3.MathSphericalPos], d3.MathSphericalPos),
    (type[d4.FourVector], d4.FourVector),
    (type[dn.PoincarePolarVector], dn.PoincarePolarVector),
    (type[dn.CartesianPosND], dn.CartesianPosND),
    # Velocities
    (type[d1.CartesianVel1D], d1.CartesianVel1D, AbstractPos),
    (type[d1.RadialVel], d1.RadialVel, AbstractPos),
    (type[d1.CartesianVel1D], d1.CartesianVel1D),  # q not needed
    (type[d1.RadialVel], d1.RadialVel),  # q not needed
    (type[d2.CartesianVel2D], d2.CartesianVel2D, d2.AbstractPos2D),
    (type[d2.CartesianVel2D], d2.CartesianVel2D),  # q not needed
    (type[d2.PolarVel], d2.PolarVel, d2.AbstractPos2D),
    (type[d2.PolarVel], d2.PolarVel),
    (type[d2.TwoSphereVel], d2.TwoSphereVel, d2.AbstractPos2D),
    (type[d2.TwoSphereVel], d2.TwoSphereVel),  # q not needed
    (type[d3.CartesianVel3D], d3.CartesianVel3D, AbstractPos),
    (type[d3.CartesianVel3D], d3.CartesianVel3D),  # q not needed
    (type[d3.CylindricalVel], d3.CylindricalVel, AbstractPos),
    (type[d3.CylindricalVel], d3.CylindricalVel),  # q not needed
    (type[d3.SphericalVel], d3.SphericalVel, AbstractPos),
    (type[d3.SphericalVel], d3.SphericalVel),  # q not needed
    (type[d3.LonLatSphericalVel], d3.LonLatSphericalVel, AbstractPos),
    (type[d3.LonLatSphericalVel], d3.LonLatSphericalVel),  # q not needed
    (type[d3.LonCosLatSphericalVel], d3.LonCosLatSphericalVel, AbstractPos),
    (type[d3.LonCosLatSphericalVel], d3.LonCosLatSphericalVel),  # q not needed
    (type[d3.MathSphericalVel], d3.MathSphericalVel, AbstractPos),
    (type[d3.MathSphericalVel], d3.MathSphericalVel),  # q not needed
    (type[d3.ProlateSpheroidalVel], d3.ProlateSpheroidalVel, AbstractPos),
    (type[d3.ProlateSpheroidalVel], d3.ProlateSpheroidalVel),  # q not needed
    (type[dn.CartesianVelND], dn.CartesianVelND, dn.AbstractPosND),
    (type[dn.CartesianVelND], dn.CartesianVelND),  # q not needed
    # Accelerations
    (type[d1.CartesianAcc1D], d1.CartesianAcc1D, AbstractVel, AbstractPos),
    (type[d1.RadialAcc], d1.RadialAcc, AbstractVel, AbstractPos),
    (type[d1.CartesianAcc1D], d1.CartesianAcc1D),  # q,p not needed
    (type[d1.RadialAcc], d1.RadialAcc),  # q,p not needed
    (type[d2.CartesianAcc2D], d2.CartesianAcc2D, d2.AbstractVel2D, d2.AbstractPos2D),
    (type[d2.CartesianAcc2D], d2.CartesianAcc2D),  # q,p not needed
    (type[d3.CartesianAcc3D], d3.CartesianAcc3D, AbstractVel, AbstractPos),
    (type[d3.CartesianAcc3D], d3.CartesianAcc3D),  # q,p not needed
    (type[d3.CylindricalAcc], d3.CylindricalAcc, AbstractVel, AbstractPos),
    (type[d3.CylindricalAcc], d3.CylindricalAcc),  # q,p not needed
    (type[d3.SphericalAcc], d3.SphericalAcc, AbstractVel, AbstractPos),
    (type[d3.SphericalAcc], d3.SphericalAcc),  # q,p not needed
    (type[d3.LonLatSphericalAcc], d3.LonLatSphericalAcc, AbstractVel, AbstractPos),
    (type[d3.LonLatSphericalAcc], d3.LonLatSphericalAcc),  # q,p not needed
    (type[d3.MathSphericalAcc], d3.MathSphericalAcc, AbstractVel, AbstractPos),
    (type[d3.MathSphericalAcc], d3.MathSphericalAcc),  # q,p not needed
    (type[dn.CartesianAccND], dn.CartesianAccND, dn.AbstractVelND, dn.AbstractPosND),
    (type[dn.CartesianAccND], dn.CartesianAccND),  # q,p not needed
)
def vconvert(
    target: type[AbstractVector], current: AbstractVector, /, *args: Any, **kw: Any
) -> AbstractVector:
    """Self transforms.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    ## 1D:

    >>> q = cxv.CartesianPos1D.from_(1, "m")
    >>> cxv.vconvert(cxv.CartesianPos1D, q) is q
    True

    >>> q = cxv.RadialPos.from_(1, "m")
    >>> cxv.vconvert(cxv.RadialPos, q) is q
    True

    >>> p = cxv.CartesianVel1D.from_(1, "m/s")
    >>> cxv.vconvert(cxv.CartesianVel1D, p) is p
    True
    >>> cxv.vconvert(cxv.CartesianVel1D, p, q) is p
    True

    >>> p = cxv.RadialVel.from_(1, "m/s")
    >>> cxv.vconvert(cxv.RadialVel, p) is p
    True
    >>> cxv.vconvert(cxv.RadialVel, p, q) is p
    True

    >>> a = cxv.CartesianAcc1D.from_(1, "m/s2")
    >>> cxv.vconvert(cxv.CartesianAcc1D, a) is a
    True
    >>> cxv.vconvert(cxv.CartesianAcc1D, a, p, q) is a
    True

    >>> a = cxv.RadialAcc.from_(1, "m/s2")
    >>> cxv.vconvert(cxv.RadialAcc, a) is a
    True
    >>> cxv.vconvert(cxv.RadialAcc, a, p, q) is a
    True

    ## 2D:

    >>> q = cxv.CartesianPos2D.from_([1, 2], "m")
    >>> cxv.vconvert(cxv.CartesianPos2D, q) is q
    True

    >>> q = cxv.PolarPos(r=u.Quantity(1, "m"), phi=u.Quantity(10, "deg"))
    >>> cxv.vconvert(cxv.PolarPos, q) is q
    True

    >>> q = cxv.TwoSpherePos(theta=u.Quantity(10, "deg"), phi=u.Quantity(14, "deg"))
    >>> cxv.vconvert(cxv.TwoSpherePos, q) is q
    True

    >>> p = cxv.CartesianVel2D.from_([1, 2], "m/s")
    >>> cxv.vconvert(cxv.CartesianVel2D, p) is p
    True

    >>> p = cxv.PolarVel(r=u.Quantity(1, "m/s"), phi=u.Quantity(10, "deg/s"))
    >>> cxv.vconvert(cxv.PolarVel, p) is p
    True

    >>> a = cxv.CartesianAcc2D.from_([1, 2], "m/s2")
    >>> cxv.vconvert(cxv.CartesianAcc2D, a) is a
    True

    ## 3D:

    Cartesian to Cartesian:

    >>> vec = cxv.CartesianPos3D.from_([1, 2, 3], "km")
    >>> cxv.vconvert(cxv.CartesianPos3D, vec) is vec
    True

    Cylindrical to Cylindrical:

    >>> vec = cxv.CylindricalPos(rho=u.Quantity(1, "km"),
    ...                          phi=u.Quantity(2, "deg"),
    ...                          z=u.Quantity(3, "km"))
    >>> cxv.vconvert(cxv.CylindricalPos, vec) is vec
    True

    Spherical to Spherical:

    >>> vec = cxv.SphericalPos(r=u.Quantity(1, "km"),
    ...                        theta=u.Quantity(2, "deg"),
    ...                        phi=u.Quantity(3, "deg"))
    >>> cxv.vconvert(cxv.SphericalPos, vec) is vec
    True

    LonLatSpherical to LonLatSpherical:

    >>> vec = cxv.LonLatSphericalPos(lon=u.Quantity(1, "deg"),
    ...                              lat=u.Quantity(2, "deg"),
    ...                              distance=u.Quantity(3, "km"))
    >>> cxv.vconvert(cxv.LonLatSphericalPos, vec) is vec
    True

    MathSpherical to MathSpherical:

    >>> vec = cxv.MathSphericalPos(r=u.Quantity(1, "km"),
    ...                            theta=u.Quantity(2, "deg"),
    ...                            phi=u.Quantity(3, "deg"))
    >>> cxv.vconvert(cxv.MathSphericalPos, vec) is vec
    True

    For these transformations the position does not matter since the
    self-transform returns the velocity unchanged.

    >>> vec = cxv.CartesianPos3D.from_([1, 2, 3], "km")

    Cartesian to Cartesian velocity:

    >>> dif = cxv.CartesianVel3D.from_([1, 2, 3], "km/s")
    >>> cxv.vconvert(cxv.CartesianVel3D, dif, vec) is dif
    True

    Cylindrical to Cylindrical velocity:

    >>> dif = cxv.CylindricalVel(rho=u.Quantity(1, "km/s"),
    ...                          phi=u.Quantity(2, "mas/yr"),
    ...                          z=u.Quantity(3, "km/s"))
    >>> cxv.vconvert(cxv.CylindricalVel, dif, vec) is dif
    True

    Spherical to Spherical velocity:

    >>> dif = cxv.SphericalVel(r=u.Quantity(1, "km/s"),
    ...                        theta=u.Quantity(2, "mas/yr"),
    ...                        phi=u.Quantity(3, "mas/yr"))
    >>> cxv.vconvert(cxv.SphericalVel, dif, vec) is dif
    True

    LonLatSpherical to LonLatSpherical velocity:

    >>> dif = cxv.LonLatSphericalVel(lon=u.Quantity(1, "mas/yr"),
    ...                              lat=u.Quantity(2, "mas/yr"),
    ...                              distance=u.Quantity(3, "km/s"))
    >>> cxv.vconvert(cxv.LonLatSphericalVel, dif, vec) is dif
    True

    LonCosLatSpherical to LonCosLatSpherical velocity:

    >>> dif = cxv.LonCosLatSphericalVel(lon_coslat=u.Quantity(1, "mas/yr"),
    ...                                 lat=u.Quantity(2, "mas/yr"),
    ...                                 distance=u.Quantity(3, "km/s"))
    >>> cxv.vconvert(cxv.LonCosLatSphericalVel, dif, vec) is dif
    True

    MathSpherical to MathSpherical velocity:

    >>> dif = cxv.MathSphericalVel(r=u.Quantity(1, "km/s"),
    ...                            theta=u.Quantity(2, "mas/yr"),
    ...                            phi=u.Quantity(3, "mas/yr"))
    >>> cxv.vconvert(cxv.MathSphericalVel, dif, vec) is dif
    True

    Similarly for the these accelerations:

    >>> a = cxv.CartesianAcc3D.from_([1, 1, 1], "m/s2")
    >>> cxv.vconvert(cxv.CartesianAcc3D, a) is a
    True

    ## N-D:

    >>> x = cxv.CartesianPosND.from_([1, 2, 3, 4], "km")
    >>> cxv.vconvert(cxv.CartesianPosND, x) is x
    True

    >>> v = cxv.CartesianVelND.from_([1, 2, 3, 4], "km/s")
    >>> cxv.vconvert(cxv.CartesianVelND, v, x) is v
    True
    >>> cxv.vconvert(cxv.CartesianVelND, v) is v
    True

    >>> a = cxv.CartesianAccND.from_([1, 2, 3, 4], "m/s2")
    >>> cxv.vconvert(cxv.CartesianAccND, a, v, x) is a
    True
    >>> cxv.vconvert(cxv.CartesianAccND, a) is a
    True

    """
    return current


# =============================================================================
# Generic Convert


@dispatch
def vconvert(
    to_vector: type[AbstractPos],
    from_vector: type[AbstractPos],
    params: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """AbstractPos -> CartesianPos1D -> AbstractPos."""
    params, aux = vconvert(
        from_vector.cartesian_type,
        from_vector,
        params,
        in_aux=in_aux,
        out_aux=None,
        units=units,
    )
    params, aux = vconvert(
        to_vector.cartesian_type,
        from_vector.cartesian_type,
        params,
        in_aux=aux,
        out_aux=aux,
        units=units,
    )
    params, aux = vconvert(
        to_vector,
        to_vector.cartesian_type,
        params,
        in_aux=aux,
        out_aux=out_aux,
        units=units,
    )
    return params, aux


# =============================================================================
# Dimension drop


@dispatch.multi(
    (type[d1.CartesianPos1D], type[d2.CartesianPos2D], ct.ParamsDict),
    (type[d1.CartesianPos1D], type[d3.CartesianPos3D], ct.ParamsDict),
    (type[d1.RadialPos], type[d2.PolarPos], ct.ParamsDict),
    (type[d2.CartesianPos2D], type[d3.CartesianPos3D], ct.ParamsDict),
    (type[d1.RadialPos], type[d3.SphericalPos], ct.ParamsDict),
    (type[d1.RadialPos], type[d3.MathSphericalPos], ct.ParamsDict),
)
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[AbstractPos],
    from_vector: type[AbstractPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CartesianPos2D -> CartesianPos1D.

    The `y` and `z` coordinates are dropped.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    ## 2D:

    >>> params = {"x": 1, "y": 2}

    >>> with warnings.catch_warnings(action="ignore"):
    ...     cxv.vconvert(cxv.CartesianPos1D, cxv.CartesianPos2D, params)
    ({'x': 1}, {})

    >>> x = cxv.CartesianPos2D.from_([1, 2], "km")
    >>> with warnings.catch_warnings(action="ignore"):
    ...     y = cxv.vconvert(cxv.CartesianPos1D, x)
    >>> print(y)
    <CartesianPos1D: (x) [km]
        [1]>

    >>> x = cxv.PolarPos(r=u.Quantity(1, "km"), phi=u.Quantity(10, "deg"))

    >>> with warnings.catch_warnings(action="ignore"):
    ...     y = cxv.vconvert(cxv.RadialPos, x)
    >>> print(y)
    <RadialPos: (r) [km]
        [1]>

    ## 3D:

    >>> params = {"x": 1, "y": 2, "z": 3}

    >>> with warnings.catch_warnings(action="ignore"):
    ...     cxv.vconvert(cxv.CartesianPos1D, cxv.CartesianPos3D, params)
    ({'x': 1}, {})

    >>> x = cxv.CartesianPos3D.from_([1.0, 2.0, 3.0], "km")
    >>> with warnings.catch_warnings(action="ignore"):
    ...     y = cxv.vconvert(cxv.CartesianPos1D, x)
    >>> print(y)
    <CartesianPos1D: (x) [km]
        [1.]>

    >>> with warnings.catch_warnings(action="ignore"):
    ...     cxv.vconvert(cxv.CartesianPos2D, cxv.CartesianPos3D, params)
    ({'x': 1, 'y': 2}, {})

    >>> x = cxv.CartesianPos3D.from_([1.0, 2.0, 3.0], "km")

    >>> with warnings.catch_warnings(action="ignore"):
    ...     y = cxv.vconvert(cxv.CartesianPos2D, x)
    >>> print(y)
    <CartesianPos2D: (x, y) [km]
        [1. 2.]>

    >>> params = {"r": 1, "theta": 14, "phi": 10}
    >>> usys = u.unitsystem("km", "deg")

    >>> with warnings.catch_warnings(action="ignore"):
    ...     cxv.vconvert(cxv.RadialPos, cxv.SphericalPos,
    ...                       params, units=usys)
    ({'r': Array(1, dtype=int32, ...)}, {})

    >>> x = cxv.SphericalPos(r=u.Quantity(1, "km"),
    ...                     theta=u.Quantity(14, "deg"),
    ...                     phi=u.Quantity(10, "deg"))
    >>> with warnings.catch_warnings(action="ignore"):
    ...     y = cxv.vconvert(cxv.RadialPos, x)
    >>> print(y)
    <RadialPos: (r) [km]
        [1]>

    >>> params = {"r": 1, "theta": 10, "phi": 14}
    >>> usys = u.unitsystem("km", "deg")

    >>> with warnings.catch_warnings(action="ignore"):
    ...     cxv.vconvert(cxv.RadialPos, cxv.MathSphericalPos,
    ...                  params, units=usys)
    ({'r': Array(1, dtype=int32, ...)}, {})

    >>> x = cxv.MathSphericalPos(r=u.Quantity(1, "km"),
    ...                          theta=u.Quantity(10, "deg"),
    ...                          phi=u.Quantity(14, "deg"))

    >>> with warnings.catch_warnings(action="ignore"):
    ...     y = cxv.vconvert(cxv.RadialPos, x)
    >>> print(y)
    <RadialPos: (r) [km]
        [1]>

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return {k: p[k] for k in to_vector.components}, (out_aux or {})



#!/usr/bin/env -S uv run --script  # noqa: EXE001
# /// script
#    dependencies = ["nox", "nox_uv"]
# ///
"""Nox setup."""

import argparse
import shutil
from pathlib import Path

import nox
from nox_uv import session

nox.needs_version = ">=2024.3.2"
nox.options.default_venv_backend = "uv"

DIR = Path(__file__).parent.resolve()

# =============================================================================
# Linting


@session(uv_groups=["lint"], reuse_venv=True, default=True)
def lint(s: nox.Session, /) -> None:
    """Run the linter."""
    s.notify("precommit")
    s.notify("pylint")


@session(uv_groups=["lint"], reuse_venv=True)
def precommit(s: nox.Session, /) -> None:
    """Run the linter."""
    s.run("pre-commit", "run", "--all-files", *s.posargs)


@session(uv_groups=["lint"], reuse_venv=True)
def pylint(s: nox.Session, /) -> None:
    """Run PyLint."""
    s.run("pylint", "coordinax", *s.posargs)


# =============================================================================
# Testing


@session(uv_groups=["test"], reuse_venv=True, default=True)
def test(s: nox.Session, /) -> None:
    """Run the unit and regular tests."""
    s.notify("pytest", posargs=s.posargs)
    # s.notify("benchmark", posargs=s.posargs)


@session(uv_groups=["test"], reuse_venv=True)
def pytest(s: nox.Session, /) -> None:
    """Run the unit and regular tests."""
    s.run("pytest", *s.posargs)


# @session(uv_groups=["test"], reuse_venv=True)
# def benchmark(s: nox.Session, /) -> None:
#     """Run the benchmarks."""
#     s.run("pytest", "tests/benchmark", "--codspeed", *s.posargs)


# =============================================================================
# Documentation


@session(uv_groups=["docs"], reuse_venv=True)
def docs(s: nox.Session, /) -> None:
    """Build the docs. Pass "--serve" to serve. Pass "-b linkcheck" to check links."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--serve", action="store_true", help="Serve after building")
    parser.add_argument(
        "-b",
        dest="builder",
        default="html",
        help="Build target (default: html)",
    )
    parser.add_argument("--output-dir", dest="output_dir", default="_build")
    args, posargs = parser.parse_known_args(s.posargs)

    if args.builder != "html" and args.serve:
        s.error("Must not specify non-HTML builder with --serve")

    s.chdir("docs")

    if args.builder == "linkcheck":
        s.run(
            "sphinx-build",
            "-b",
            "linkcheck",
            ".",
            "_build/linkcheck",
            *posargs,
        )
        return

    shared_args = (
        "-n",  # nitpicky mode
        "-T",  # full tracebacks
        f"-b={args.builder}",
        f"-d {args.output_dir}/doctrees",
        "-D language=en",
        ".",
        f"{args.output_dir}/{args.builder}",
        *posargs,
    )

    if args.serve:
        s.run("sphinx-autobuild", *shared_args)
    else:
        s.run("sphinx-build", "--keep-going", *shared_args)


@session(uv_groups=["docs"], reuse_venv=True)
def build_api_docs(s: nox.Session, /) -> None:
    """Build (regenerate) API docs."""
    s.chdir("docs")
    s.run(
        "sphinx-apidoc",
        "-o",
        "api/",
        "--module-first",
        "--no-toc",
        "--force",
        "../src/coordinax",
    )


# =============================================================================


@session(uv_groups=["build"], reuse_venv=True)
def build(s: nox.Session, /) -> None:
    """Build an SDist and wheel."""
    build_path = DIR.joinpath("build")
    if build_path.exists():
        shutil.rmtree(build_path)
    s.run("python", "-m", "build")


################################################################################

if __name__ == "__main__":
    nox.main()



"""Bases."""

__all__ = (
    "AbstractVectorLike",
    "AbstractVector",
    "AbstractCartesian",
    # Type Guards
    "is_vectorlike",
    "is_vector",
    # Flags
    "AttrFilter",
    # Utils
    "VectorAttribute",
    "ToUnitsOptions",
)

from .attribute import VectorAttribute
from .base import AbstractVectorLike, is_vectorlike
from .cartesian import AbstractCartesian
from .flags import AttrFilter
from .register_unxt import ToUnitsOptions
from .vector import AbstractVector, is_vector

# Register by import
# isort: split
from . import (
    register_constructors,  # noqa: F401
    register_dataclassish,  # noqa: F401
    register_primitives,  # noqa: F401
    register_unxt,  # noqa: F401
)



"""Intra-ecosystem Compatibility."""

__all__: tuple[str, ...] = ()


from collections.abc import Callable
from typing import TypeVar, Union

import equinox as eqx
from jaxtyping import Shaped
from jaxtyping._array_types import _MetaAbstractArray
from plum import conversion_method as _conversion_method, convert, dispatch

import quaxed.numpy as jnp
import unxt as u
from dataclassish import field_values
from unxt.quantity import BareQuantity

from .api import vector
from coordinax._src.vectors.base import AbstractVector
from coordinax._src.vectors.d1 import (
    CartesianAcc1D,
    CartesianPos1D,
    CartesianVel1D,
    RadialAcc,
    RadialVel,
)
from coordinax._src.vectors.d2 import CartesianAcc2D, CartesianPos2D, CartesianVel2D
from coordinax._src.vectors.d3 import CartesianAcc3D, CartesianPos3D, CartesianVel3D
from coordinax._src.vectors.dn import CartesianAccND, CartesianPosND, CartesianVelND
from coordinax._src.vectors.utils import full_shaped

T = TypeVar("T")


def conversion_method(type_from: type, type_to: type) -> Callable[[T], T]:
    """Typed version of conversion method."""
    return _conversion_method(type_from, type_to)  # type: ignore[return-value]


#####################################################################
# TODO: move to unxt


@dispatch(precedence=1)
def dimension_of(obj: _MetaAbstractArray, /) -> u.dims.AbstractDimension:
    """Get the dimension of a jaxtyping-annotated object."""
    return u.dimension_of(obj.array_type)


@dispatch
def dimension_of(obj: Union, /) -> u.dims.AbstractDimension:
    # Call dimension_of on all the __args__ of the Union and make sure they are
    # all the same
    dims = {dimension_of(arg) for arg in obj.__args__}
    dims = eqx.error_if(dims, len(dims) != 1, "Cannot mix dimensions in a Union.")
    return dims.pop()


#####################################################################
# Construct from Quantity


class Dim:
    """Dimension enumeration."""

    LENGTH = u.dimension("length")
    SPEED = u.dimension("speed")
    ACCELERATION = u.dimension("acceleration")


@dispatch
def vector(q: u.AbstractQuantity, /) -> AbstractVector:  # noqa: C901
    """Construct a vector from a quantity.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> print(cx.vecs.vector(u.Quantity(1, "km")))
    <CartesianPos1D: (x) [km]
        [1]>

    >>> print(cx.vecs.vector(u.Quantity([1], "km")))
    <CartesianPos1D: (x) [km]
        [1]>

    >>> print(cx.vecs.vector(u.Quantity(1, "km/s")))
    <CartesianVel1D: (x) [km / s]
        [1]>

    >>> print(cx.vecs.vector(u.Quantity([1], "km/s")))
    <CartesianVel1D: (x) [km / s]
        [1]>

    >>> print(cx.vecs.vector(u.Quantity(1, "km/s2")))
    <CartesianAcc1D: (x) [km / s2]
        [1]>

    >>> print(cx.vecs.vector(u.Quantity([1], "km/s2")))
    <CartesianAcc1D: (x) [km / s2]
        [1]>

    >>> print(cx.vecs.vector(u.Quantity([1, 2], "km")))
    <CartesianPos2D: (x, y) [km]
        [1 2]>

    >>> print(cx.vecs.vector(u.Quantity([1, 2], "km/s")))
    <CartesianVel2D: (x, y) [km / s]
        [1 2]>

    >>> print(cx.vecs.vector(u.Quantity([1, 2], "km/s2")))
    <CartesianAcc2D: (x, y) [km / s2]
        [1 2]>

    >>> print(cx.vecs.vector(u.Quantity([1, 2, 3], "km")))
    <CartesianPos3D: (x, y, z) [km]
        [1 2 3]>

    >>> print(cx.vecs.vector(u.Quantity([1, 2, 3], "km/s")))
    <CartesianVel3D: (x, y, z) [km / s]
        [1 2 3]>

    >>> print(cx.vecs.vector(u.Quantity([1, 2, 3], "km/s2")))
    <CartesianAcc3D: (x, y, z) [km / s2]
        [1 2 3]>

    >>> print(cx.vecs.vector(u.Quantity([0, 1, 2, 3], "km")))
    <CartesianPosND: (q) [km]
        [[0]
         [1]
         [2]
         [3]]>

    >>> print(cx.vecs.vector(u.Quantity([0, 1, 2, 3], "km/s")))
    <CartesianVelND: (q) [km / s]
        [[0]
         [1]
         [2]
         [3]]>

    >>> print(cx.vecs.vector(u.Quantity([0, 1, 2, 3], "km/s2")))
    <CartesianAccND: (q) [km / s2]
        [[0]
         [1]
         [2]
         [3]]>

    >>> try: print(cx.vecs.vector(u.Quantity([1], "Msun")))
    ... except ValueError as e: print(e)
    Cannot construct a Cartesian vector from Quantity['mass']([1], unit='solMass').

    """
    # TODO: use dispatch instead for these matches
    match (u.dimension_of(q), jnp.atleast_1d(q).shape[-1]):
        case (Dim.LENGTH, 0) | (Dim.LENGTH, 1):
            return vector(CartesianPos1D, q)
        case (Dim.SPEED, 0) | (Dim.SPEED, 1):
            return vector(CartesianVel1D, q)
        case (Dim.ACCELERATION, 0) | (Dim.ACCELERATION, 1):
            return vector(CartesianAcc1D, q)
        case (Dim.LENGTH, 2):
            return vector(CartesianPos2D, q)
        case (Dim.SPEED, 2):
            return vector(CartesianVel2D, q)
        case (Dim.ACCELERATION, 2):
            return vector(CartesianAcc2D, q)
        case (Dim.LENGTH, 3):
            return vector(CartesianPos3D, q)
        case (Dim.SPEED, 3):
            return vector(CartesianVel3D, q)
        case (Dim.ACCELERATION, 3):
            return vector(CartesianAcc3D, q)
        case (Dim.LENGTH, _):
            return vector(CartesianPosND, q)
        case (Dim.SPEED, _):
            return vector(CartesianVelND, q)
        case (Dim.ACCELERATION, _):
            return vector(CartesianAccND, q)
        case _:
            msg = f"Cannot construct a Cartesian vector from {q}."
            raise ValueError(msg)


@dispatch
def vector(q: list[float | int], unit: str, /) -> AbstractVector:
    """Construct a vector from a list and unit.

    Examples
    --------
    >>> import coordinax as cx

    >>> print(cx.vecs.vector([1], "km"))
    <CartesianPos1D: (x) [km]
        [1]>

    >>> print(cx.vecs.vector([1], "km/s"))
    <CartesianVel1D: (x) [km / s]
        [1]>

    >>> print(cx.vecs.vector([1], "km/s2"))
    <CartesianAcc1D: (x) [km / s2]
        [1]>

    >>> print(cx.vecs.vector([1, 2], "km"))
    <CartesianPos2D: (x, y) [km]
        [1 2]>

    >>> print(cx.vecs.vector([1, 2], "km/s"))
    <CartesianVel2D: (x, y) [km / s]
        [1 2]>

    >>> print(cx.vecs.vector([1, 2], "km/s2"))
    <CartesianAcc2D: (x, y) [km / s2]
        [1 2]>

    >>> print(cx.vecs.vector([1, 2, 3], "km"))
    <CartesianPos3D: (x, y, z) [km]
        [1 2 3]>

    >>> print(cx.vecs.vector([1, 2, 3], "km/s"))
    <CartesianVel3D: (x, y, z) [km / s]
        [1 2 3]>

    >>> print(cx.vecs.vector([1, 2, 3], "km/s2"))
    <CartesianAcc3D: (x, y, z) [km / s2]
        [1 2 3]>

    """
    return vector(u.Quantity.from_(jnp.array(q), unit))


#####################################################################
# Convert to Quantity


def _vec_diff_to_q(obj: AbstractVector, /) -> u.AbstractQuantity:
    """`coordinax.AbstractVector` -> `unxt.u.AbstractQuantity`."""
    return jnp.stack(tuple(field_values(full_shaped(obj))), axis=-1)


@conversion_method(type_from=RadialAcc, type_to=BareQuantity)
@conversion_method(type_from=RadialVel, type_to=BareQuantity)
@conversion_method(type_from=CartesianAcc1D, type_to=BareQuantity)
@conversion_method(type_from=CartesianVel1D, type_to=BareQuantity)
@conversion_method(type_from=CartesianAcc2D, type_to=BareQuantity)
@conversion_method(type_from=CartesianVel2D, type_to=BareQuantity)
@conversion_method(type_from=CartesianAcc3D, type_to=BareQuantity)
@conversion_method(type_from=CartesianVel3D, type_to=BareQuantity)
def vec_diff_to_uncheckedq(obj: AbstractVector, /) -> Shaped[BareQuantity, "*batch N"]:
    """Differentials -> `unxt.BareQuantity`.

    Examples
    --------
    >>> from plum import convert
    >>> from unxt.quantity import BareQuantity
    >>> import coordinax as cx

    ## 1D

    >>> cart_vel = cx.vecs.CartesianVel1D.from_([1], "km/s")
    >>> convert(cart_vel, BareQuantity)
    BareQuantity(Array([1], dtype=int32), unit='km / s')

    >>> cart_acc = cx.vecs.CartesianAcc1D.from_([1], "km/s2")
    >>> convert(cart_acc, BareQuantity)
    BareQuantity(Array([1], dtype=int32), unit='km / s2')

    >>> rad_vel = cx.vecs.RadialVel.from_([1], "km/s")
    >>> convert(rad_vel, BareQuantity)
    BareQuantity(Array([1], dtype=int32), unit='km / s')

    >>> rad_acc = cx.vecs.RadialAcc.from_([1], "km/s2")
    >>> convert(rad_acc, BareQuantity)
    BareQuantity(Array([1], dtype=int32), unit='km / s2')

    ## 2D

    >>> vel = cx.vecs.CartesianVel2D.from_([1, 2], "km/s")
    >>> convert(vel, BareQuantity)
    BareQuantity(Array([1, 2], dtype=int32), unit='km / s')

    >>> acc = cx.vecs.CartesianAcc2D.from_([1, 2], "km/s2")
    >>> convert(acc, BareQuantity)
    BareQuantity(Array([1, 2], dtype=int32), unit='km / s2')

    # 3D

    >>> vel = cx.CartesianVel3D.from_([1, 2, 3], "km/s")
    >>> convert(vel, BareQuantity)
    BareQuantity(Array([1, 2, 3], dtype=int32), unit='km / s')

    >>> acc = cx.vecs.CartesianAcc3D.from_([1, 2, 3], "km/s2")
    >>> convert(acc, BareQuantity)
    BareQuantity(Array([1, 2, 3], dtype=int32), unit='km / s2')

    """
    return convert(_vec_diff_to_q(obj), BareQuantity)


@conversion_method(type_from=RadialAcc, type_to=u.Quantity)
@conversion_method(type_from=RadialVel, type_to=u.Quantity)
@conversion_method(type_from=CartesianAcc1D, type_to=u.Quantity)
@conversion_method(type_from=CartesianVel1D, type_to=u.Quantity)
@conversion_method(type_from=CartesianAcc2D, type_to=u.Quantity)
@conversion_method(type_from=CartesianVel2D, type_to=u.Quantity)
@conversion_method(type_from=CartesianAcc3D, type_to=u.Quantity)
@conversion_method(type_from=CartesianVel3D, type_to=u.Quantity)
def vec_diff_to_q(obj: AbstractVector, /) -> Shaped[u.Quantity, "*batch N"]:
    """1D Differentials -> `unxt.Quantity`.

    Examples
    --------
    >>> from plum import convert
    >>> import unxt as u
    >>> import coordinax as cx

    ## 1D

    >>> cart_vel = cx.vecs.CartesianVel1D.from_([1], "km/s")
    >>> convert(cart_vel, u.Quantity)
    Quantity(Array([1], dtype=int32), unit='km / s')

    >>> cart_acc = cx.vecs.CartesianAcc1D.from_([1], "km/s2")
    >>> convert(cart_acc, u.Quantity)
    Quantity(Array([1], dtype=int32), unit='km / s2')

    >>> rad_vel = cx.vecs.RadialVel.from_([1], "km/s")
    >>> convert(rad_vel, u.Quantity)
    Quantity(Array([1], dtype=int32), unit='km / s')

    >>> rad_acc = cx.vecs.RadialAcc.from_([1], "km/s2")
    >>> convert(rad_acc, u.Quantity)
    Quantity(Array([1], dtype=int32), unit='km / s2')

    ## 2D

    >>> vel = cx.vecs.CartesianVel2D.from_([1, 2], "km/s")
    >>> convert(vel, u.Quantity)
    Quantity(Array([1, 2], dtype=int32), unit='km / s')

    >>> acc = cx.vecs.CartesianAcc2D.from_([1, 2], "km/s2")
    >>> convert(acc, u.Quantity)
    Quantity(Array([1, 2], dtype=int32), unit='km / s2')

    # 3D

    >>> vel = cx.CartesianVel3D.from_([1, 2, 3], "km/s")
    >>> convert(vel, u.Quantity)
    Quantity(Array([1, 2, 3], dtype=int32), unit='km / s')

    >>> acc = cx.vecs.CartesianAcc3D.from_([1, 2, 3], "km/s2")
    >>> convert(acc, u.Quantity)
    Quantity(Array([1, 2, 3], dtype=int32), unit='km / s2')

    """
    return convert(_vec_diff_to_q(obj), u.Quantity)
